#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Standard library imports.
import os

# Enthought library imports.
from pyface.api import confirm, error, YES
from traits.api import Bool, Dict, HasTraits, provides, Instance, \
        List, Password, Property, Unicode
from traitsui.api import Handler, Item, View
from traitsui.menu import Action, OKCancelButtons

# Local imports.
from apptools.permissions.i_user import IUser
from .i_user_database import IUserDatabase
from .i_user_storage import IUserStorage, UserStorageError
from .select_user import select_user


@provides(IUser)
class _LoginUser(HasTraits):
    """This represents the login data and view."""

    #### '_LoginUser' interface ###############################################

    # The user name.
    name = Unicode

    # The user password.
    password = Password

    # The default view.
    traits_view = View(Item(name='name'), Item(name='password'),
            title="Login", kind='modal', buttons=OKCancelButtons)


class _ChangePassword(HasTraits):
    """This represents the change password data and view."""

    #### '_ChangePassword' interface ##########################################

    # The user name.
    name = Unicode

    # The new user password.
    new_password = Password

    # The confirmed new user password.
    confirm_new_password = Password

    # The default view.
    traits_view = View(Item(name='name', style='readonly'),
            Item(name='new_password'), Item(name='confirm_new_password'),
            title="Change password", kind='modal', buttons=OKCancelButtons)


class _ViewUserAccount(HasTraits):
    """This represents a single account when in a view."""

    #### '_ViewUserAccount' interface #########################################

    # The name the user uses to identify themselves.
    name = Unicode

    # A description of the user (typically their full name).
    description = Unicode

    # The password
    password = Password

    # The password confirmation.
    confirm_password = Password

    # The user database.
    user_db = Instance(IUserDatabase)


class _UserAccountHandler(Handler):
    """The base traits handler for user account views."""

    ###########################################################################
    # 'Handler' interface.
    ###########################################################################

    def close(self, info, is_ok):
        """Reimplemented to validate the data."""

        # If the user cancelled then close the dialog.
        if not is_ok:
            return True

        # Validate the form, only closing the dialog if the form is valid.
        return self.validate(self._user_account(info))

    ###########################################################################
    # '_UserAccountHandler' interface.
    ###########################################################################

    def validate(self, vuac):
        """Validate the given object and return True if there were no problems.
        """

        if not vuac.name.strip():
            self.error("A user name must be given.")
            return False

        return True

    @staticmethod
    def error(msg):
        """Display an error message to the user."""

        error(None, msg)

    ###########################################################################
    # Trait handlers.
    ###########################################################################

    def _search_clicked(self, info):
        """Invoked by the "Search" button."""

        # Get the user name.
        vuac = self._user_account(info)

        name, description = vuac.user_db._select_user(vuac.name)
        if not name:
            return

        # Update the viewed object.
        vuac.name = name
        vuac.description = description
        vuac.password = vuac.confirm_password = ''

    ###########################################################################
    # Private interface.
    ###########################################################################

    @staticmethod
    def _user_account(info):
        """Return the user account instance being handled."""

        return info.ui.context['object']


class _AddUserAccountHandler(_UserAccountHandler):
    """The traits handler for the add user account view."""

    ###########################################################################
    # '_UserAccountHandler' interface.
    ###########################################################################

    def validate(self, vuac):
        """Validate the given object and return True if there were no problems.
        """

        if not super(_AddUserAccountHandler, self).validate(vuac):
            return False

        if not _validate_password(vuac.password, vuac.confirm_password):
            return False

        return True


class _UserAccountView(View):
    """The base view for handling user accounts."""

    #### 'View' interface #####################################################

    kind = 'modal'

    ###########################################################################
    # Trait handlers.
    ###########################################################################

    def _buttons_default(self):
        """Return the view's buttons."""

        # Create an action that will search the database.
        buttons = [Action(name="Search")]
        buttons.extend(OKCancelButtons)

        return buttons


class _AddUserAccountView(_UserAccountView):
    """A view to handle adding a user account."""

    #### 'View' interface #####################################################

    title = "Add a user"

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, **traits):
        """Initialise the object."""

        super(_AddUserAccountView, self).__init__(Item(name='name'),
                Item(name='description'), Item(name='password'),
                Item(name='confirm_password'), **traits)


class _ModifyUserAccountView(_AddUserAccountView):
    """A view to handle modifying a user account."""

    #### 'View' interface #####################################################

    title = "Modify a user"


class _DeleteUserAccountView(_UserAccountView):
    """A view to handle deleting a user account."""

    #### 'View' interface #####################################################

    title = "Delete a user"

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, **traits):
        """Initialise the object."""

        super(_DeleteUserAccountView, self).__init__(Item(name='name'),
                Item(name='description', style='readonly'), **traits)


class User(HasTraits):
    """The user implementation.  We don't store any extra information other
    than that defined by IUser."""



    #### 'IUser' interface ####################################################

    name = Unicode

    authenticated = Bool(False)

    description = Unicode

    blob = Dict


@provides(IUserDatabase)
class UserDatabase(HasTraits):
    """This implements a user database that supports IUser for the default user
    manager (ie. using password authorisation) except that it leaves the actual
    access of the data to an implementation of IUserStorage."""



    #### 'IUserDatabase' interface ############################################

    can_change_password = Property

    can_add_user = Property

    can_modify_user = Property

    can_delete_user = Property

    user_storage = Instance(IUserStorage)

    #### Private interface ####################################################

    # Set if updating the user blob internally.
    _updating_blob_internally = Bool(False)

    ###########################################################################
    # 'IUserDatabase' interface.
    ###########################################################################

    def bootstrapping(self):
        """See if we are bootstrapping."""

        try:
            bootstrap = self.user_storage.is_empty()
        except UserStorageError:
            # Suppress the error and assume it isn't empty.
            bootstrap = False

        return bootstrap

    def authenticate_user(self, user):
        """Authenticate a user."""

        # Get the login details.
        name = user.name
        lu = _LoginUser(name=name)

        if not lu.edit_traits().result:
            return False

        # Get the user account and compare passwords.
        try:
            name, description, blob = self.user_storage.authenticate_user(
                    lu.name.strip(), lu.password)
        except UserStorageError as e:
            self._us_error(e)
            return False

        # Update the user details.
        user.name = name
        user.description = description

        # Suppress the trait notification.
        self._updating_blob_internally = True
        user.blob = blob
        self._updating_blob_internally = False

        return True

    def unauthenticate_user(self, user):
        """Unauthenticate a user."""

        return self.user_storage.unauthenticate_user(user)

    def change_password(self, user):
        """Change a user's password."""

        # Get the new password.
        name = user.name
        np = _ChangePassword(name=name)

        if not np.edit_traits().result:
            return

        # Validate the password.
        if not _validate_password(np.new_password, np.confirm_new_password):
            return

        # Update the password in the database.
        try:
            self.user_storage.update_password(name, np.new_password)
        except UserStorageError as e:
            self._us_error(e)

    def add_user(self):
        """Add a user."""

        # Get the data from the user.
        vuac = _ViewUserAccount(user_db=self)
        view = _AddUserAccountView()
        handler = _AddUserAccountHandler()

        if vuac.edit_traits(view=view, handler=handler).result:
            # Add the data to the database.
            try:
                self.user_storage.add_user(vuac.name.strip(), vuac.description,
                        vuac.password)
            except UserStorageError as e:
                self._us_error(e)

    def modify_user(self):
        """Modify a user."""

        # Get the data from the user.
        vuac = _ViewUserAccount(user_db=self)
        view = _ModifyUserAccountView()
        handler = _UserAccountHandler()

        if vuac.edit_traits(view=view, handler=handler).result:
            # Update the data in the database.
            try:
                self.user_storage.modify_user(vuac.name.strip(),
                        vuac.description, vuac.password)
            except UserStorageError as e:
                self._us_error(e)

    def delete_user(self):
        """Delete a user."""

        # Get the data from the user.
        vuac = _ViewUserAccount(user_db=self)
        view = _DeleteUserAccountView()
        handler = _UserAccountHandler()

        if vuac.edit_traits(view=view, handler=handler).result:
            # Make absolutely sure.
            name = vuac.name.strip()

            if confirm(None, "Are you sure you want to delete the user \"%s\"?" % name) == YES:
                # Delete the data from the database.
                try:
                    self.user_storage.delete_user(name)
                except UserStorageError as e:
                    self._us_error(e)

    def matching_user(self, name):
        """Select a user."""

        name, description = self._select_user(name)
        if not name:
            return None

        return User(name=name, description=description)

    def user_factory(self):
        """Create a new user object."""

        user = User(name=os.environ.get('USER', ''))

        # Monitor when the blob changes.
        user.on_trait_change(self._blob_changed, name='blob')

        return user

    ###########################################################################
    # Trait handlers.
    ###########################################################################

    def _get_can_change_password(self):
        """See if a user can change their password."""

        return 'user_password' in self.user_storage.capabilities

    def _get_can_add_user(self):
        """See if a user can be added."""

        return 'user_add' in self.user_storage.capabilities

    def _get_can_modify_user(self):
        """See if a user can be modified."""

        return 'user_modify' in self.user_storage.capabilities

    def _get_can_delete_user(self):
        """See if a user can be deleted."""

        return 'user_delete' in self.user_storage.capabilities

    def _user_storage_default(self):
        """Return the default storage for the user data."""

        # Defer to an external storage manager if there is one.
        try:
            from apptools.permissions.external.user_storage import UserStorage
        except ImportError:
            from apptools.permissions.default.user_storage import UserStorage

        return UserStorage()

    def _blob_changed(self, user, tname, old, new):
        """Invoked when the user's blob data changes."""

        if not self._updating_blob_internally:
            try:
                self.user_storage.update_blob(user.name, user.blob)
            except UserStorageError as e:
                self._us_error(e)

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _select_user(self, name):
        """Select a user returning the data as a tuple of name and description.
        """

        # Get all users that satisfy the criteria.
        try:
            users = self.user_storage.matching_users(name)
        except UserStorageError as e:
            self._us_error(e)
            return '', ''

        if len(users) == 0:
            error(None, "There is no user that matches \"%s\"." % name)
            return '', ''

        return select_user(users)

    @staticmethod
    def _us_error(e):
        """Display a message to the user after a UserStorageError exception has
        been raised.  If the message is empty then we assume the user has
        already been informed."""

        msg = str(e)
        if msg:
            error(None, msg)


def _validate_password(password, confirmation):
    """Validate a password and return True if it is valid."""

    MIN_PASSWORD_LEN = 6

    if password != confirmation:
        error(None, "The passwords do not match.")
        return False

    if not password:
        error(None, "A password must be given.")
        return False

    if len(password) < MIN_PASSWORD_LEN:
        error(None, "The password must be at least %d characters long." % MIN_PASSWORD_LEN)
        return False

    return True



""" Abstract base class for all adapter factories. """

# Standard library imports.
import logging

# Enthought library imports.
from traits.api import Delegate, HasTraits, Instance

# Local imports.
from .adapter_manager import AdapterManager


logger = logging.getLogger(__name__)


class AbstractAdapterFactory(HasTraits):
    """ Abstract base class for all adapter factories.

    Adapter factories define behavioural extensions for classes.

    """

    #### 'AbstractAdapterFactory' interface ###################################

    # The adapter manager that the factory is registered with (this will be
    # None iff the factory is not registered with a manager).
    adapter_manager = Instance(AdapterManager)

    # The type system used by the factory (it determines 'is_a' relationships
    # and type MROs etc). By default we use standard Python semantics.
    type_system = Delegate('adapter_manager')

    ###########################################################################
    # 'AbstractAdapterFactory' interface.
    ###########################################################################

    def adapt(self, adaptee, target_class, *args, **kw):
        """ Returns an adapter that adapts an object to the target class.

        Returns None if the factory cannot produce such an adapter.

        """

        if self._can_adapt(adaptee, target_class, *args, **kw):
            adapter = self._adapt(adaptee, target_class, *args, **kw)
            if adapter is None:
                logger.warn(self._get_warning_message(adaptee, target_class))

        else:
            adapter = None

        return adapter

    ###########################################################################
    # Protected 'AbstractAdapterFactory' interface.
    ###########################################################################

    def _can_adapt(self, adaptee, target_class, *args, **kw):
        """ Returns True if the factory can produce an appropriate adapter. """

        raise NotImplementedError

    def _adapt(self, adaptee, target_class, *args, **kw):
        """ Returns an adapter that adapts an object to the target class. """

        raise NotImplementedError

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_warning_message(self, adaptee, target_class):
        """ Returns a warning message.

        The warning message is used when a factory fails to adapt something
        that it said it could!

        """

        message = '%s failed to adapt %s to %s' % (
            self.__class__.__name__,
            str(adaptee),
            target_class.__name__
        )

        return message

#### EOF ######################################################################



#!/usr/bin/env python
# -*- coding: utf-8 -*-



from nose.tools import assert_equal

from ..lru_cache import LRUCache


def test_cache_callback():
    dropped = []
    callback = lambda *args: dropped.append(args)
    c = LRUCache(2, cache_drop_callback=callback)
    for i in range(5):
        c[i] = str(i)

    expected = [(0, '0'), (1, '1'), (2, '2')]
    assert_equal(expected, dropped)


def test_cache_len():
    c = LRUCache(2)
    assert_equal(0, len(c))
    assert_equal(2, c.size)
    c[0] = None
    assert_equal(1, len(c))
    assert_equal(2, c.size)
    c[1] = None
    assert_equal(2, len(c))
    assert_equal(2, c.size)
    c[2] = None
    assert_equal(2, len(c))
    assert_equal(2, c.size)


def test_cache_resize():
    c = LRUCache(1)
    c[0] = 0
    c[1] = 1

    assert_equal(1, len(c))
    assert_equal(1, c.size)
    assert 0 not in c
    assert 1 in c

    c.size = 3
    assert_equal(1, len(c))
    assert_equal(3, c.size)
    assert 1 in c

    c[2] = 2
    assert_equal(2, len(c))
    assert_equal(3, c.size)
    assert 1 in c
    assert 2 in c

    c[3] = 3
    assert_equal(3, len(c))
    assert_equal(3, c.size)
    assert 1 in c
    assert 2 in c
    assert 3 in c

    c[4] = 4
    assert_equal(3, len(c))
    assert_equal(3, c.size)
    assert 1 not in c
    assert 2 in c
    assert 3 in c
    assert 4 in c


def test_cache_items():
    c = LRUCache(2)
    assert_equal([], list(c.items()))

    c[0] = str(0)
    c[1] = str(1)
    c[2] = str(2)

    expected = [(1, '1'), (2, '2')]
    assert_equal(expected, sorted(c.items()))


def test_cache_idempotency():
    c = LRUCache(2)
    c[1] = 1
    assert_equal(1, len(c))
    c[1] = 1
    assert_equal(1, len(c))

    c[2] = 2
    assert_equal(2, len(c))
    c[1] = 1
    assert_equal(2, len(c))
    c[2] = 2
    assert_equal(2, len(c))

    c[3] = 3
    assert_equal(2, len(c))
    c[3] = 3
    assert_equal(2, len(c))


def test_cache_keys_values():
    c = LRUCache(2)
    assert_equal([], list(c.items()))

    c[0] = str(0)
    c[1] = str(1)
    c[2] = str(2)

    expected = [1, 2]
    assert_equal(expected, sorted(c.keys()))
    assert_equal(list(map(str, expected)), sorted(c.values()))


def test_cache_clear():
    c = LRUCache(2)
    for i in range(c.size):
        c[i] = i
    assert_equal(c.size, len(c))
    c.clear()
    assert_equal(0, len(c))


def test_cache_get():
    c = LRUCache(2)
    for i in range(c.size):
        c[i] = i
    assert_equal(0, c.get(0))
    assert_equal(None, c.get(c.size))


def test_updated_event():
    c = LRUCache(2)
    events = []

    # Traits can't handle builtins as handlers
    c.on_trait_change(lambda x: events.append(x), 'updated')

    c[0] = 0
    print(list(c.keys()))
    assert_equal(sorted(events), [[0]])

    c[1] = 1
    assert_equal(sorted(events), [[0], [0, 1]])

    c[2] = 2
    assert_equal(sorted(events), [[0], [0, 1], [1, 2]])

    # Getting items doesn't fire anything
    c[1]
    assert_equal(sorted(events), [[0], [0, 1], [1, 2]])

    c.get(3)
    assert_equal(sorted(events), [[0], [0, 1], [1, 2]])



"""
Globals for the scripting package.
"""
# Author: Prabhu Ramachandran <prabhu@aero.iitb.ac.in>
# Copyright (c) 2008, Prabhu Ramachandran and Enthought, Inc.
# License: BSD Style.


# The global recorder.
_recorder = None

def get_recorder():
    """Return the global recorder.  Does not create a new one if none
    exists.
    """
    global _recorder
    return _recorder

def set_recorder(rec):
    """Set the global recorder instance.
    """
    global _recorder
    _recorder = rec





# Copyright (c) 2007-2011 by Enthought, Inc.
# All rights reserved.
""" Supports limiting access to parts of an application to authorised users.
    Part of the AppTools project of the Enthought Tool Suite.
"""



from .abstract_adapter_factory import AbstractAdapterFactory
from .abstract_factory import AbstractFactory
from .abstract_type_system import AbstractTypeSystem
from .adaptable import Adaptable
from .adapter import Adapter
from .adapter_factory import AdapterFactory
from .adapter_manager import AdapterManager
from .factory import Factory
from .hook import add_pre, add_post, remove_pre, remove_post
from .python_type_system import PythonObject, PythonTypeSystem
from .type_manager import TypeManager



#-------------------------------------------------------------------------------
#
#  An abstract base class implementation of the ITemplateDataNameItem interface
#  that looks for specified sub-contexts in its input context and if one match
#  is found, outputs that context; otherwise if more than one match is found it
#  outputs a context containing all matching sub-contexts found.
#
#  Written by: David C. Morrill
#
#  Date: 07/29/2007
#
#  (c) Copyright 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" An abstract base class implementation of the ITemplateDataNameItem interface
    that looks for specified sub-contexts in its input context and if one match
    is found, outputs that context; otherwise if more than one match is found it
    outputs a context containing all matching sub-contexts found.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from traits.api \
    import HasPrivateTraits, Instance, Property, provides, on_trait_change

from apptools.template.itemplate_data_context \
    import ITemplateDataContext

from apptools.template.itemplate_data_name_item \
    import ITemplateDataNameItem

from apptools.template.template_impl \
    import Template

from .template_data_context \
    import TemplateDataContext

#-------------------------------------------------------------------------------
#  'AnyContextDataNameItem' class:
#-------------------------------------------------------------------------------

class AnyContextDataNameItem ( Template ):
    """ An abstract base class implementation of the ITemplateDataNameItem
        interface that looks for specified sub-contexts in its input context
        and if one match is found, outputs that context; otherwise if more than
        one match is found it outputs a context containing all matching
        sub-contexts found.
    """

    implements ( ITemplateDataNameItem )

    #-- 'ITemplateDataNameItem' Interface Implementation -----------------------

    # The data context which this data name item should match against:
    input_data_context = Instance( ITemplateDataContext )

    # The data context containing the data values and/or contexts this data
    # name item matches:
    output_data_context = Instance( ITemplateDataContext )

    # The ITemplateChoice instance representing the current settings of the
    # data name item. This value must be read/write, and must be overridden by
    # sublasses.
    data_name_item_choice = Property

    # The alternative choices the user has for the data name item settings for
    # the current input data context. The list may be empty, in which case the
    # user cannot change the settings of the data name item. This value can be
    # read only, and must be overridden by subclasses.
    data_name_item_choices = Property

    #-- Private Traits ---------------------------------------------------------

    # The current input data context:
    current_input_data_context = Property

    #-- Partially Abstract Methods (Can be overridden in subclasses) -----------

    def filter ( self, name, context ):
        """ Returns **True** if the specified *context* called *name* should be
            included in the output context; and **False** otherwise.
        """
        return False

    #-- Property Implementations -----------------------------------------------

    def _get_data_name_item_choice ( self ):
        raise NotImplementedError

    def _set_data_name_item_choice ( self, value ):
        raise NotImplementedError

    def _get_data_name_item_choices ( self ):
        raise NotImplementedError

    def _get_current_input_data_context ( self ):
        return self.input_data_context

    #-- Trait Event Handlers ---------------------------------------------------

    def _input_data_context_changed ( self ):
        """ Handles the 'input_data_context' trait being changed.
        """
        self.inputs_changed()

    #-- Private Methods --------------------------------------------------------

    def inputs_changed ( self ):
        """ Handles any of the input values being changed. May be called by
            subclasses.
        """
        output_context = None
        input_context  = self.input_data_context
        if input_context is not None:
            contexts = {}

            # Process each name/context in the input data contexts, and only add
            # those that match the subclass's filter to the output context:
            filter = self.filter
            gdc    = input_context.get_data_context
            for name in input_context.data_contexts:
                if list(filter( name, gdc( name ) )):
                    contexts[ name ] = context

            # If the result set is not empty, create an output context for it:
            n = len( contexts )
            if n == 1:
                output_context = list(values.values())[0]
            elif n > 1:
                output_context = TemplateDataContext(
                    data_context_path = input_context.data_context_path,
                    data_context_name = input_context.data_context_name,
                    contexts          = contexts )

        # Set the new output context:
        self.output_data_context = output_context




#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" The event fired by the tree model when it changes. """


# Enthought library imports.
from traits.api import HasTraits, Instance

# Local imports.
from .binding import Binding


# Classes for event traits.
class NamingEvent(HasTraits):
    """ Information about tree model changes. """

    # The old binding.
    old_binding = Instance(Binding)

    # The new binding.
    new_binding = Instance(Binding)

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought util package component>
#------------------------------------------------------------------------------
""" Implementation of hooked methods. """


def add_pre(klass, method_name, callable):
    """ Adds a pre-hook to method 'method_name' on class 'klass'. """

    _add_hook(klass, method_name, callable, pre=True)

    return

def add_post(klass, method_name, callable):
    """ Adds a post-hook to method 'method_name' on class 'klass'. """

    _add_hook(klass, method_name, callable, pre=False)

    return

def remove_pre(klass, method_name, callable):
    """ Removes a pre-hook from method 'method_name' on class 'klass'. """

    _remove_hook(klass, method_name, callable, pre=True)

    return

def remove_post(klass, method_name, callable):
    """ Removes a post-hook from method 'method_name' on class 'klass'. """

    _remove_hook(klass, method_name, callable, pre=False)

    return

###############################################################################
# Private functions.
###############################################################################

def _add_hook(klass, method_name, callable, pre):
    """ Adds a pre/post hook to method 'method_name' on class 'klass'. """

    # Get the method on the klass.
    method = getattr(klass, method_name)

    # Have we already hooked it?
    if hasattr(method, '__pre__'):
        hooked_method = method

    # Obviously not!
    else:
        def hooked_method(self, *args, **kw):
            for pre in hooked_method.__pre__:
                pre(self, *args, **kw)

            result = method(self, *args, **kw)

            for post in hooked_method.__post__:
                result = post(self, result, *args, **kw)

            return result

        # Python < 2.4 does not allow this.
        try:
            hooked_method.__name__ = method_name

        except:
            pass

        hooked_method.__pre__  = []
        hooked_method.__post__ = []

        # Is the original method actually defined on the class, or is it
        # inherited?
        hooked_method.__inherited__ = method_name not in klass.__dict__

        # Save the original method...
        #
        # fixme: Twisted uses 'method.im_func' instead of 'method' here, but
        # both seem to work just as well!
        setattr(klass, '__hooked__' + method_name, method)

        # ... and put in the hooked one!
        setattr(klass, method_name, hooked_method)

    if pre:
        hooked_method.__pre__.append(callable)

    else:
        hooked_method.__post__.append(callable)

    return

def _remove_hook(klass, method_name, callable, pre):
    """ Removes a pre/post hook from method 'method_name' on class 'klass'. """

    # Get the method on the klass.
    method = klass.__dict__[method_name]

    # Is it actually hooked?
    if hasattr(method, '__pre__'):
        # Remove the hook.
        if pre:
            method.__pre__.remove(callable)

        else:
            method.__post__.remove(callable)

        # If there are no more hooks left then cleanup.
        if len(method.__pre__) == 0 and len(method.__post__) == 0:
            # If the method is inherited then just delete the hooked version.
            if method.__inherited__:
                delattr(klass, method_name)

            # Otherwise, reinstate the original method.
            else:
                original = getattr(klass, '__hooked__' + method_name)
                setattr(klass, method_name, original)

            # Remove the saved original method.
            delattr(klass, '__hooked__' + method_name)

    return

#### EOF ######################################################################



#-------------------------------------------------------------------------------
#
#  A feature-based Traits UI plug-in for viewing templates.
#
#  Written by: David C. Morrill
#
#  Date: 07/30/2007
#
#  (c) Copyright 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" A feature-based Traits UI plug-in for viewing templates.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import sys

from os.path \
    import split, splitext

from pickle \
    import dump, load

from traits.api \
    import HasPrivateTraits, Str, Instance, Any, File, Button, on_trait_change

from traitsui.api \
    import View, VGroup, HGroup, Tabbed, Item, InstanceEditor, Theme, Label

from traitsui.wx.themed_text_editor \
    import ThemedTextEditor

from etsdevtools.developer.features.api \
    import DropFile

from pyface.api \
    import FileDialog, OK

from etsdevtools.developer.helper.themes \
    import TButton

from blockcanvas.app.utils \
    import import_log_files

from apptools.template.api \
    import ITemplate, ITemplateDataContext, TemplateDataNames, Template

from .enable_editor \
    import EnableEditor

#-- Adapters that might be used ------------------------------------------------

from apptools.template.impl.base_data_context_adapter \
    import BaseDataContextAdapter

#-------------------------------------------------------------------------------
#  Helper class:
#-------------------------------------------------------------------------------

class Message ( HasPrivateTraits ):

    #-- Trait Definitions ------------------------------------------------------

    # The message to be displayed:
    message = Str

    #-- Traits View Definitions ------------------------------------------------

    view = View(
        Item( 'message',
              style      = 'readonly',
              show_label = False,
              editor     = ThemedTextEditor( theme = '@GBB' ),
        )
    )

    #-- Object Overrides -------------------------------------------------------

    def __init__ ( self, message = '', **traits ):
        traits.setdefault( 'message', message )

        super( Message, self ).__init__( **traits )

# Define some standard messages:
no_context        = Message( 'Please provide a data context' )
no_template       = Message( 'Please provide a view template' )
no_template_found = Message( 'No view templates found in Python source file' )
no_bindings       = Message( 'Please resolve all required template data bindings' )
no_options        = Message( 'No options are currently available' )

#-------------------------------------------------------------------------------
#  'TemplateView' class:
#-------------------------------------------------------------------------------

class TemplateView ( HasPrivateTraits ):
    """ A feature-based Traits UI plug-in for viewing templates.
    """

    #-- Public Traits ----------------------------------------------------------

    # The name of the plugin:
    name = Str( 'Template View' )

    # The data context supplying the data to be viewed:
    context = Instance( ITemplateDataContext,
                        connect = 'to: data context' )

    # The name of the file containing a template view:
    file_name = File( drop_file = DropFile(
                     extensions = [ '.py', '.tv', '.las' ],
                     tooltip = 'Drop a LAS data file, a saved view template '
                               'or a Python source file containing a view '
                               'template here.' ),
                     connect = 'to: template view' )

    # The template to view:
    template = Instance( ITemplate )

    #-- Private Traits ---------------------------------------------------------

    # The name of the file to save the template in:
    save_file_name = File

    # The current data names:
    data_names = Instance( TemplateDataNames )

    # The TemplateDataNames or Message being viewed:
    names_view = Any( no_context )

    # The name of the most recently loaded template file:
    template_file_name = File

    # The template or message being viewed:
    template_view = Any( no_template )

    # The options view for the currently active template:
    options_view = Any( no_options )

    # The event fired when the user wants to save the template:
    save_template = Button( 'Save Template' )

    #-- Traits View Definitions ------------------------------------------------

    view = View(
        VGroup(
            HGroup(
                Item( 'file_name', show_label = False, width = 350 ),
                TButton( 'save_template',
                      label        = 'Save Template',
                      enabled_when = 'template is not None' ),
                group_theme = Theme( '@GFB', margins = ( -7, -5 ) )
            ),
            Tabbed(
                VGroup(
                    '8',
                    Label( 'Data Bindings',
                           item_theme = Theme( '@GBB', alignment = 'center' )
                    ),
                    Item( 'names_view',
                          style      = 'custom',
                          resizable  = True,
                          editor     = InstanceEditor(),
                          export     = 'DockWindowShell',
                          item_theme = Theme( '@GFB', margins = ( -5, -1 ) )
                    ),
                    label       = 'Data Bindings',
                    show_labels = False
                ),
                Item( 'template_view',
                      label     = 'Template View',
                      style     = 'custom',
                      resizable = True,
                      editor    = InstanceEditor( view = 'template_view' ),
                      export    = 'DockWindowShell'
                ),
                Item( 'options_view',
                      label     = 'Template Options',
                      style     = 'custom',
                      resizable = True,
                      editor    = InstanceEditor( view = 'options_view' ),
                      export    = 'DockWindowShell'
                ),
                id          = 'tabbed',
                dock        = 'horizontal',
                show_labels = False
            ),
        ),
        id = 'template.test.template_view.TemplateView'
    )

    #-- Trait Event Handlers ---------------------------------------------------

    def _context_changed ( self, context ):
        """ Handles the 'context' trait being changed.
        """
        if context is None:
            self.names_view = no_context
        elif self.template is None:
            self.names_view = no_template
        else:
            self._create_view()

    def _template_changed ( self, template ):
        """ Handles the 'template' trait being changed.
        """
        if self.context is None:
            self.template_view = no_context
        else:
            self._create_view()

    def _file_name_changed ( self, file_name ):
        """ Handles the 'file_name' trait being changed.
        """
        ext = splitext( file_name )[1]
        if ext == '.py':
            self._load_python_template( file_name )
        elif ext == '.tv':
            self._load_pickled_template( file_name )
        elif ext == '.las':
            self._load_las_file( file_name )
        else:
            # fixme: Display an informational message here...
            pass

    def _save_template_changed ( self ):
        """ Handles the user clicking the 'Save Template' button.
        """
        self._save_pickled_template()

    @on_trait_change( 'data_names.unresolved_data_names' )
    def _on_unresolved_data_names ( self ):
        if len( self.data_names.unresolved_data_names ) == 0:
            self._create_object_view()
        elif not isinstance( self.template_view, Message ):
            self.template_view = no_bindings
            self.options_view  = no_options

    @on_trait_change( 'template.template_mutated?' )
    def _on_template_mutated ( self ):
        """ Handles a mutable template changing.
        """
        if self.context is not None:
            self._create_view()

    #-- Private Methods --------------------------------------------------------

    def _load_python_template ( self, file_name ):
        """ Attempts to load a template from a Python source file.
        """
        path, name    = split( file_name )
        sys.path[0:0] = [ path ]
        try:
            ###values = {}
            module_name, ext = splitext( name )
            module = __import__( module_name )
            values = module.__dict__
            ###execfile( file_name, values )
            template = values.get( 'template' )
            if template is None:
                templates = []
                for value in list(values.values()):
                    try:
                        if (issubclass( value, Template ) and
                            ###(value.__module__ == '__builtin__')):
                            (value.__module__ == module_name)):
                            templates.append( value )
                    except:
                        pass

                for i, template in enumerate( templates ):
                    for t in templates[ i + 1: ]:
                        if issubclass( template, t ):
                            break
                    else:
                        break
                else:
                    self.template_view = no_template_found
                    return

            if not isinstance( template, Template ):
                template = template()

            self.template           = template
            self.template_file_name = file_name

        except Exception as excp:
            self.template_view = Message( str( excp ) )

        # Clean up the Python path:
        del sys.path[0]

    def _load_pickled_template ( self, file_name ):
        """ Attempts to load a template from a pickle.
        """
        # fixme: Implement this...load template from .tv pickle file.
        fh     = None
        delete = False
        try:
            fh            = open( file_name, 'rb' )
            file_name     = load( fh )
            path, name    = split( file_name )
            sys.path[0:0] = [ path ]
            delete        = True
            module_name, ext = splitext( name )
            module = __import__( module_name )
            self.template = load( fh )
            self.template_file_name = file_name
        except Exception as excp:
            import traceback
            traceback.print_exc()
            self.template_view = Message( str( excp ) )

        if fh is not None:
            fh.close()

        if delete:
            del sys.path[0]

    def _load_las_file ( self, file_name ):
        """ Creates a data context from the specified LAS file.
        """
        try:
            self.context = import_log_files( file_name, 'las' )
        except Exception as excp:
            self.names_view = Message( str( excp ) )

    def _save_pickled_template ( self ):
        file_name = self.save_file_name or self.file_name
        fd        = FileDialog(
                        action       = 'save as',
                        default_path = file_name )
                        #wildcard     = 'Template files (*.tv)|*.tv|' )
        if fd.open() == OK:
            self.save_file_name = file_name = fd.path
            fh = None
            try:
                fh = open( file_name, 'wb' )
                dump( self.template_file_name, fh, -1 )
                dump( self.template.template_from_object(), fh, -1 )
            except:
                # fixme: Display an informational message here...
                import traceback
                traceback.print_exc()

            if fh is not None:
                fh.close()

    def _create_view ( self ):
        """ Begins the process of creating a live view from a template and
            data context object.
        """
        self.data_names = self.names_view = nv = TemplateDataNames(
                              context    = self.context,
                              data_names = self.template.names_from_template() )

        if len( nv.unresolved_data_names ) == 0:
            self._create_object_view()
        else:
            self.template_view = no_bindings

    def _create_object_view ( self ):
        """ Create the object view from the current template.
        """
        self.template.object_from_template()
        self.template_view = self.options_view = self.template




#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from traits.api import HasTraits, Instance, List, Unicode
from traitsui.api import Item, TableEditor, View
from traitsui.menu import OKCancelButtons
from traitsui.table_column import ObjectColumn


class _Role(HasTraits):
    """This represents the role model."""

    #### '_RoleModel' interface ###############################################

    # The role name.
    name = Unicode

    # The role description.
    description = Unicode

    # The permissions ids.
    permissions = List


class _RolesView(HasTraits):
    """This represents the view used to select a role."""

    #### '_UsersView' interface ###############################################

    # The list of roles to select from.
    model = List(_Role)

    # The selected user.
    selection = Instance(_Role)

    # The editor used by the view.
    table_editor = TableEditor(columns=[ObjectColumn(name='name'),
                    ObjectColumn(name='description')],
            selected='selection', sort_model=True, configurable=False)

    # The default view.
    traits_view = View(Item('model', show_label=False, editor=table_editor),
            title="Select a Role", style='readonly', kind='modal',
            buttons=OKCancelButtons)


def select_role(roles):
    """Return a single role from the given list of roles."""

    # Construct the model.
    model = [_Role(name=name, description=description, permissions=permissions)
            for name, description, permissions in roles]

    # Construct the view.
    view = _RolesView(model=model)

    if view.configure_traits() and view.selection is not None:
        role = view.selection.name, view.selection.description, view.selection.permissions
    else:
        role = '', '', []

    return role



from .dict_context_adapter import DictContextAdapter
from .dict_context_adapter_factory import DictContextAdapterFactory
from .instance_context_adapter import InstanceContextAdapter
from .instance_context_adapter_factory import InstanceContextAdapterFactory
from .list_context_adapter import ListContextAdapter
from .list_context_adapter_factory import ListContextAdapterFactory
from .trait_list_context_adapter import TraitListContextAdapter
from .trait_list_context_adapter_factory import TraitListContextAdapterFactory
from .tuple_context_adapter import TupleContextAdapter
from .tuple_context_adapter_factory import TupleContextAdapterFactory

from .trait_dict_context_adapter import TraitDictContextAdapter



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought undo package component>
#------------------------------------------------------------------------------



from .abstract_command import AbstractCommand
from .command_stack import CommandStack
from .i_command import ICommand
from .i_command_stack import ICommandStack
from .i_undo_manager import IUndoManager
from .undo_manager import UndoManager



# Standard library imports
from pickle import *
import sys, new
import logging
from types import GeneratorType

# Enthought library imports
from apptools.persistence.updater import __replacement_setstate__


logger = logging.getLogger(__name__)

##############################################################################
# class 'NewUnpickler'
##############################################################################
class NewUnpickler(Unpickler):
    """ An unpickler that implements a two-stage pickling process to make it
    possible to unpickle complicated Python object hierarchies where the
    unserialized state of an object depends on the state of other objects in
    the same pickle.
    """

    def load(self, max_pass=-1):
        """Read a pickled object representation from the open file.

        Return the reconstituted object hierarchy specified in the file.
        """
        # List of objects to be unpickled.
        self.objects = []

        # We overload the load_build method.
        dispatch = self.dispatch
        dispatch[BUILD[0]] = NewUnpickler.load_build

        # call the super class' method.
        ret = Unpickler.load(self)
        self.initialize(max_pass)
        self.objects = []

        # Reset the Unpickler's dispatch table.
        dispatch[BUILD[0]] = Unpickler.load_build
        return ret

    def initialize(self, max_pass):
        # List of (object, generator) tuples that initialize objects.
        generators = []

        # Execute object's initialize to setup the generators.
        for obj in self.objects:
            if hasattr(obj, '__initialize__') and \
                   callable(obj.__initialize__):
                ret = obj.__initialize__()
                if isinstance(ret, GeneratorType):
                    generators.append((obj, ret))
                elif ret is not None:
                    raise UnpicklingError('Unexpected return value from '
                        '__initialize__.  %s returned %s' % (obj, ret))

        # Ensure a maximum number of passes
        if max_pass < 0:
            max_pass = len(generators)

        # Now run the generators.
        count = 0
        while len(generators) > 0:
            count += 1
            if count > max_pass:
                not_done = [x[0] for x in generators]
                msg = """Reached maximum pass count %s.  You may have
                         a deadlock!  The following objects are
                         uninitialized: %s""" % (max_pass, not_done)
                raise UnpicklingError(msg)
            for o, g in generators[:]:
                try:
                    next(g)
                except StopIteration:
                    generators.remove((o, g))

    # Make this a class method since dispatch is a class variable.
    # Otherwise, supposing the initial sweet_pickle.load call (which would
    # have overloaded the load_build method) makes a pickle.load call at some
    # point, we would have the dispatch still pointing to
    # NewPickler.load_build whereas the object being passed in will be an
    # Unpickler instance, causing a TypeError.
    def load_build(cls, obj):
        # Just save the instance in the list of objects.
        if isinstance(obj, NewUnpickler):
            obj.objects.append(obj.stack[-2])
        Unpickler.load_build(obj)
    load_build = classmethod(load_build)


class VersionedUnpickler(NewUnpickler):
    """ This class reads in a pickled file created at revision version 'n'
    and then applies the transforms specified in the updater class to
    generate a new set of objects which are at revision version 'n+1'.

    I decided to keep the loading of the updater out of this generic class
    because we will want updaters to be generated for each plugin's type
    of project.

    This ensures that the VersionedUnpickler can remain ignorant about the
    actual version numbers - all it needs to do is upgrade one release.
    """


    def __init__(self, file, updater=None):
        Unpickler.__init__(self, file)
        self.updater = updater
        return


    def find_class(self, module, name):
        """ Overridden method from Unpickler.

        NB  __setstate__ is not called until later.
        """

        if self.updater:
            # check to see if this class needs to be mapped to a new class
            # or module name
            original_module, original_name  = module, name
            #logger.debug('omodule:%s oname:%s' % (original_module, original_name))
            module, name = self.updater.get_latest(module, name)
            #logger.debug('module:%s name:%s' % (module, name))

            # load the class...
            '''__import__(module)
            mod = sys.modules[module]
            klass = getattr(mod, name)'''
            klass = self.import_name(module, name)

            # add the updater....  TODO - why the old name?
            self.add_updater(original_module, original_name, klass)

        else:
            # there is no updater so we will be reading in an up to date
            # version of the file...
            try:
                klass = Unpickler.find_class(self, module, name)
            except:
                logger.error("Looking for [%s] [%s]" % (module, name))
                logger.exception('Problem using default unpickle functionality')

            # restore the original __setstate__ if necessary
            fn = getattr(klass, '__setstate_original__', False)
            if fn:
                m = new.instancemethod(fn, None, klass)
                setattr(klass, '__setstate__', m)

        return klass


    def add_updater(self, module, name, klass):
        """ If there is an updater defined for this class we will add it to the
        class as the __setstate__ method.
        """

        fn = self.updater.setstates.get((module, name), False)

        if fn:
            # move the existing __setstate__ out of the way
            self.backup_setstate(module, klass)

            # add the updater into the class
            m = new.instancemethod(fn, None, klass)
            setattr(klass, '__updater__', m)

            # hook up our __setstate__ which updates self.__dict__
            m = new.instancemethod(__replacement_setstate__, None, klass)
            setattr(klass, '__setstate__', m)

        else:
            pass
            #print 'No updater fn to worry about'

        return


    def backup_setstate(self, module, klass):
        """ If the class has a user defined __setstate__ we back it up.
        """
        if getattr(klass, '__setstate__', False):

            if getattr(klass, '__setstate_original__', False):
                # don't overwrite the original __setstate__
                name = '__setstate__%s' % self.updater.__class__
            else:
                # backup the original __setstate__ which we will restore
                # and run later when we have finished updating the class
                name = '__setstate_original__'

            #logger.debug('renaming __setstate__ to %s' % name)
            method = getattr(klass, '__setstate__')
            m = new.instancemethod(method, None, klass)
            setattr(klass, name, m)

        else:
            # the class has no __setstate__ method so do nothing
            pass

        return


    def import_name(self, module, name):
        """
        If the class is needed for the latest version of the application then
        it should presumably exist.

        If the class no longer exists then we should perhaps return
        a proxy of the class.

        If the persisted file is at v1 say and the application is at v3 then
        objects that are required for v1 and v2 do not have to exist they only
        need to be placeholders for the state during an upgrade.
        """
        #print "importing %s %s" % (name, module)
        module = __import__(module, globals(), locals(), [name])
        return vars(module)[name]

### EOF #################################################################



# -----------------------------------------------------------------------------
#
#  Copyright (c) 2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
# -----------------------------------------------------------------------------

from traits.testing.unittest_tools import unittest
from contextlib import contextmanager
from nose.tools import assert_equal

from apptools.undo.api import CommandStack, UndoManager
from apptools.undo.tests.testing_commands import SimpleCommand, UnnamedCommand


class TestCommandStack(unittest.TestCase):
    def setUp(self):
        self.stack = CommandStack()
        undo_manager = UndoManager()
        self.stack.undo_manager = undo_manager

        self.command = SimpleCommand()

    # Command pushing tests ---------------------------------------------------

    def test_empty_command_stack(self):
        with assert_n_commands_pushed(self.stack, 0):
            pass

    def test_1_command_pushed(self):
        with assert_n_commands_pushed(self.stack, 1):
            self.stack.push(self.command)

    def test_n_command_pushed(self):
        n = 4
        with assert_n_commands_pushed(self.stack, n):
            for i in range(n):
                self.stack.push(self.command)

    # Undo/Redo tests ---------------------------------------------------------

    def test_undo_1_command(self):
        with assert_n_commands_pushed_and_undone(self.stack, 1):
            self.stack.push(self.command)
            self.assertEqual(self.stack.undo_name, self.command.name)
            self.stack.undo()

    def test_undo_n_command(self):
        n = 4
        with assert_n_commands_pushed_and_undone(self.stack, n):
            for i in range(n):
                self.stack.push(self.command)

            for i in range(n):
                self.stack.undo()

    def test_undo_unnamed_command(self):
        unnamed_command = UnnamedCommand()
        with assert_n_commands_pushed(self.stack, 1):
            self.stack.push(unnamed_command)

            # But the command cannot be undone because it has no name
            self.assertEqual(self.stack.undo_name, "")
            # This is a no-op
            self.stack.undo()

    def test_undo_redo_1_command(self):
        with assert_n_commands_pushed(self.stack, 1):
            self.stack.push(self.command)
            self.stack.undo()
            self.stack.redo()

    # Macro tests -------------------------------------------------------------

    def test_define_macro(self):
        with assert_n_commands_pushed(self.stack, 1):
            add_macro(self.stack, num_commands=2)

    def test_undo_macro(self):
        with assert_n_commands_pushed_and_undone(self.stack, 1):
            # The 2 pushes are viewed as 1 command
            add_macro(self.stack, num_commands=2)
            self.stack.undo()

    # Cleanliness tests -------------------------------------------------------

    def test_empty_stack_is_clean(self):
        self.assertTrue(self.stack.clean)

    def test_non_empty_stack_is_dirty(self):
        self.stack.push(self.command)
        self.assertFalse(self.stack.clean)

    def test_make_clean(self):
        # This makes it dirty by default
        self.stack.push(self.command)
        # Make the current tip of the stack clean
        self.stack.clean = True
        self.assertTrue(self.stack.clean)

    def test_make_dirty(self):
        # Start from a clean state:
        self.stack.push(self.command)
        self.stack.clean = True

        self.stack.clean = False
        self.assertFalse(self.stack.clean)

    def test_save_push_undo_is_clean(self):
        self.stack.push(self.command)

        self.stack.clean = True
        self.stack.push(self.command)
        self.stack.undo()
        self.assertTrue(self.stack.clean)

    def test_save_push_save_undo_is_clean(self):
        self.stack.push(self.command)

        self.stack.clean = True
        self.stack.push(self.command)
        self.stack.clean = True
        self.stack.undo()
        self.assertTrue(self.stack.clean)

    def test_push_undo_save_redo_is_dirty(self):
        self.stack.push(self.command)
        self.stack.undo()
        self.stack.clean = True
        self.stack.redo()
        self.assertFalse(self.stack.clean)


def add_macro(stack, num_commands=2):
    command = SimpleCommand()
    stack.begin_macro('Increment n times')
    try:
        for i in range(num_commands):
            stack.push(command)
    finally:
        stack.end_macro()


# Assertion helpers -----------------------------------------------------------

@contextmanager
def assert_n_commands_pushed(stack, n):
    current_length = len(stack._stack)
    yield
    # N commands have been pushed...
    assert_equal(len(stack._stack), current_length+n)
    # ... and the state is at the tip of the stack...
    assert_equal(stack._index, current_length+n-1)


@contextmanager
def assert_n_commands_pushed_and_undone(stack, n):
    current_length = len(stack._stack)
    yield
    # N commands have been pushed and then reverted. The stack still
    # contains the commands...
    assert_equal(len(stack._stack), n)
    # ... but we are back to the initial (clean) state
    assert_equal(stack._index, current_length-1)



#-------------------------------------------------------------------------------
#
#  Defines an adapter from an codetools.contexts.api.IContext
#  to an ITemplateDataContext.
#
#  Written by: David C. Morrill
#  Modified by: Robert Kern
#
#  Date: 11/16/2007
#
#  (c) Copyright 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" Defines an adapter from an codetools.contexts.api.IContext to an
    ITemplateDataContext.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from traits.api \
    import Adapter, Str, List, adapts

from traits.protocols.api \
    import AdaptationError, adapt

from codetools.contexts.api \
    import IContext

from apptools.template.itemplate_data_context \
    import ITemplateDataContext, ITemplateDataContextError

from .helper \
    import path_for

#-------------------------------------------------------------------------------
#  'IContextAdapter' class:
#-------------------------------------------------------------------------------

class IContextAdapter ( Adapter ):
    """ Defines an adapter from an codetools.contexts.api.IContext
        to an ITemplateDataContext.
    """

    adapts( IContext, ITemplateDataContext )

    #-- ITemplateDataContext Interface Implementation --------------------------

    # The path to this data context (does not include the 'data_context_name'):
    data_context_path = Str

    # The name of the data context:
    data_context_name = Str

    # A list of the names of the data values in this context:
    data_context_values = List( Str )

    # The list of the names of the sub-contexts of this context:
    data_contexts = List( Str )

    def get_data_context_value ( self, name ):
        """ Returns the data value with the specified *name*. Raises a
            **ITemplateDataContextError** if *name* is not defined as a data
            value in the context.

            Parameters
            ----------
            name : A string specifying the name of the context data value to
                be returned.

            Returns
            -------
            The data value associated with *name* in the context. The type of
            the data is application dependent.

            Raises **ITemplateDataContextError** if *name* is not associated
            with a data value in the context.
        """
        try:
            if name in self.data_context_values:
                return self.adaptee[ name ]

            raise ITemplateDataContextError(
                      "No value named '%s' found." % name )
        except Exception as excp:
            raise ITemplateDataContextError( str( excp ) )

    def get_data_context ( self, name ):
        """ Returns the **ITemplateDataContext** value associated with the
            specified *name*. Raises **ITemplateDataContextError** if *name* is
            not defined as a data context in the context.

            Parameters
            ----------
            name : A string specifying the name of the data context to be
                returned.

            Returns
            -------
            The **ITemplateDataContext** associated with *name* in the context.

            Raises **ITemplateDataContextError** if *name* is not associated
            with a data context in the context.
        """
        try:
            if name in self.data_contexts:
                bdca = IContextAdapter( self.adaptee[ name ] )
                bdca.data_context_path = path_for( self.data_context_path,
                                                   self.data_context_name )
                return bdca

            raise ITemplateDataContextError(
                      "No context named '%s' found." % name )
        except Exception as excp:
            raise ITemplateDataContextError( str( excp ) )

    #-- Traits Event Handlers --------------------------------------------------

    def _adaptee_changed ( self, context ):
        """ Handles being bound to a IContext object.
        """
        self.data_context_name = context.name
        values   = []
        contexts = []
        for name in list(context.keys()):
            value = context[ name ]
            try:
                adapt( value, IContext )
            except AdaptationError:
                # Is not a subcontext.
                values.append( name )
            else:
                # Is a subcontext.
                contexts.append( name )

        self.data_context_values = values
        self.data_contexts       = contexts




#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from traits.api import HasTraits, Instance, List, Unicode

# Local imports.
from apptools.permissions.permission import Permission


class Role(HasTraits):
    """This represents a role."""

    # The role name.
    name = Unicode

    # The role description.
    description = Unicode

    # The permissions that define the role.
    permissions = List(Instance(Permission))

    def __str__(self):
        """Return a user friendly representation."""

        s = self.description
        if not s:
            s = self.name

        return s


class Assignment(HasTraits):
    """This represents the assignment of roles to a user."""

    # The user name.
    user_name = Unicode

    # The user description.
    description = Unicode

    # The list of assigned roles.
    roles = List(Instance(Role))



#-------------------------------------------------------------------------------
#
#  Defines the external API for the template package.
#
#  Written by: David C. Morrill
#
#  Date: 07/26/2007
#
#  (c) Copyright 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" Defines the external API for the template package.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from .template_traits \
    import TDataSource, TInstance, TList, TInt, TFloat, TStr, TBool, TRange, \
           TEnum, TDerived

from .itemplate \
    import ITemplate

from .imutable_template \
    import IMutableTemplate

from .itemplate_data_context \
    import ITemplateDataContext, ITemplateDataContextError

from .template_data_name \
    import TemplateDataName

from .template_data_names \
    import TemplateDataNames

from .template_impl \
    import Template

from .mutable_template \
    import MutableTemplate

from .template_choice \
    import TemplateChoice




import numpy as np
from numpy.testing import raises, assert_allclose

from ..dict_node import H5DictNode
from .utils import open_h5file, temp_h5_file, temp_file


NODE = '/dict_node'


def test_create():
    with temp_h5_file() as h5:
        h5dict = H5DictNode.add_to_h5file(h5, NODE)
        h5dict['a'] = 1
        h5dict['b'] = 2

        assert h5dict['a'] == 1
        assert h5dict['b'] == 2


def test_is_dict_node():
    with temp_h5_file() as h5:
        node = h5.create_dict(NODE, {})
        assert H5DictNode.is_dict_node(node._h5_group)


def test_is_not_dict_node():
    with temp_h5_file() as h5:
        node = h5.create_group(NODE)
        assert not H5DictNode.is_dict_node(node)
        assert not H5DictNode.is_dict_node(node._h5_group)


def test_create_with_data():
    with temp_h5_file() as h5:
        data = {'a': 10}

        h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
        assert h5dict['a'] == 10


def test_create_with_array_data():
    foo = np.arange(100)
    bar = np.arange(150)

    with temp_h5_file() as h5:
        data = {'a': 10, 'foo': foo, 'bar': bar}

        h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
        assert h5dict['a'] == 10
        assert_allclose(h5dict['foo'], foo)
        assert_allclose(h5dict['bar'], bar)


def test_load_saved_dict_node():
    with temp_file() as filename:
        # Write data to new dict node and close.
        with open_h5file(filename, 'w') as h5:
            h5dict = H5DictNode.add_to_h5file(h5, NODE)
            h5dict['a'] = 1

        # Read dict node and make sure the data was saved.
        with open_h5file(filename, mode='r+') as h5:
            h5dict = h5[NODE]
            assert h5dict['a'] == 1
            # Change data for next test
            h5dict['a'] = 2

        # Check that data is modified by the previous write.
        with open_h5file(filename) as h5:
            h5dict = h5[NODE]
            assert h5dict['a'] == 2


def test_load_saved_dict_node_with_array():
    arr = np.arange(100)
    arr1 = np.arange(200)

    with temp_file() as filename:
        # Write data to new dict node and close.
        with open_h5file(filename, 'w') as h5:
            h5dict = H5DictNode.add_to_h5file(h5, NODE)
            h5dict['arr'] = arr

        # Read dict node and make sure the data was saved.
        with open_h5file(filename, mode='r+') as h5:
            h5dict = h5[NODE]
            assert_allclose(h5dict['arr'], arr)
            # Change data for next test
            h5dict['arr'] = arr1
            h5dict['arr_old'] = arr

        # Check that data is modified by the previous write.
        with open_h5file(filename) as h5:
            h5dict = h5[NODE]
            assert_allclose(h5dict['arr'], arr1)
            assert_allclose(h5dict['arr_old'], arr)
            # Make sure that arrays come back as arrays
            assert isinstance(h5dict['arr'], np.ndarray)
            assert isinstance(h5dict['arr_old'], np.ndarray)


def test_keys():
    with temp_h5_file() as h5:
        keys = set(('hello', 'world', 'baz1'))
        data = dict((n, 1) for n in keys)
        h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
        assert set(h5dict.keys()) == keys


def test_delete_item():
    with temp_h5_file() as h5:
        data = dict(a=10)
        h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
        del h5dict['a']
        assert 'a' not in h5dict


def test_delete_array():
    with temp_h5_file() as h5:
        data = dict(a=np.arange(10))
        h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
        del h5dict['a']
        assert 'a' not in h5dict
        assert 'a' not in h5[NODE]


def test_auto_flush():
    with temp_h5_file() as h5:
        data = dict(a=1)
        h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
        # Overwrite existing data, which should get written to disk.
        new_data = dict(b=2)
        h5dict.data = new_data
        # Load data from disk to check that data was automatically flushed.
        h5dict_from_disk = h5[NODE]
        assert 'a' not in h5dict_from_disk
        assert h5dict_from_disk['b'] == 2


def test_auto_flush_off():
    with temp_h5_file() as h5:
        data = dict(a=1)
        h5dict = H5DictNode.add_to_h5file(h5, NODE, data, auto_flush=False)
        # Overwrite existing data, but don't write to disk.
        new_data = dict(b=2)
        h5dict.data = new_data
        # Load data from disk to check that it's unchanged.
        h5dict_from_disk = h5[NODE]
        assert h5dict_from_disk['a'] == 1
        assert 'b' not in h5dict_from_disk
        # Manually flush, and check that data was written
        h5dict.flush()
        h5dict_from_disk = h5[NODE]
        assert 'a' not in h5dict_from_disk
        assert h5dict_from_disk['b'] == 2


@raises(KeyError)
def test_undefined_key():
    with temp_h5_file() as h5:
        data = dict(a='int')
        h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
        del h5dict['b']


def test_basic_dtypes():
    with temp_h5_file() as h5:
        data = dict(a_int=1, a_float=1.0, a_str='abc')
        h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
        assert isinstance(h5dict['a_int'], int)
        assert isinstance(h5dict['a_float'], float)
        assert isinstance(h5dict['a_str'], str)


def test_mixed_type_list():
    with temp_h5_file() as h5:
        data = dict(a=[1, 1.0, 'abc'])
        h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
        for value, dtype in zip(h5dict['a'], (int, float, str)):
            assert isinstance(value, dtype)


def test_dict():
    with temp_h5_file() as h5:
        data = dict(a=dict(b=1, c=2))
        h5dict = H5DictNode.add_to_h5file(h5, NODE, data)
        sub_dict = h5dict['a']
        assert sub_dict['b'] == 1
        assert sub_dict['c'] == 2


@raises(AssertionError)
def test_wrap_self_raises_error():
    with temp_h5_file() as h5:
        H5DictNode.add_to_h5file(h5, NODE)
        node = h5[NODE]
        H5DictNode(node)


if __name__ == '__main__':
    from numpy import testing
    testing.run_module_suite()



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2006 by Enthought, Inc.
#  All rights reserved.
#
#  Author: Dave Peterson <dpeterson@enthought.com>
#
#-----------------------------------------------------------------------------

# Standard library imports
import logging

# Enthought library imports
import apptools.sweet_pickle as sweet_pickle
from apptools.sweet_pickle.global_registry import _clear_global_registry
from traits.api import Bool, Float, HasTraits, Int, Str


logger = logging.getLogger(__name__)


##############################################################################
# Classes to use within the tests
##############################################################################

class Foo(HasTraits):
    _enthought_pickle_version = Int(1)
    b1 = Bool(False)
    f1 = Float(1)
    i1 = Int(1)
    s1 = Str('foo')

class Bar(HasTraits):
    _enthought_pickle_version = Int(2)
    b2 = Bool(True)
    f2 = Float(2)
    i2 = Int(2)
    s2 = Str('bar')

class Baz(HasTraits):
    _enthought_pickle_version = Int(3)
    b3 = Bool(False)
    f3 = Float(3)
    i3 = Int(3)
    s3 = Str('baz')
    def __setstate__(self, state):
        logger.debug('Running Baz\'s original __setstate__')
        if state['_enthought_pickle_version'] < 3:
            info = [('b2', 'b3'), ('f2', 'f3'), ('i2', 'i3'), ('s2', 's3')]
            for old, new in info:
                if old in state:
                    state[new] = state[old]
                    del state[old]
            state['_enthought_pickle_version'] = 3
        self.__dict__.update(state)

### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought undo package component>
#------------------------------------------------------------------------------



# Enthought library imports.
from traits.api import Bool, Event, Instance, Int, Interface, Unicode


class IUndoManager(Interface):
    """ The undo manager interface.  An undo manager is responsible for one or
    more command stacks.  Typically an application would have a single undo
    manager.
    """

    #### 'IUndoManager' interface #############################################

    # This is the currently active command stack and may be None.  Typically it
    # is set when some sort of editor becomes active.
    active_stack = Instance('apptools.undo.api.ICommandStack')

    # This reflects the clean state of the currently active command stack.  It
    # is intended to support a "document modified" indicator in the GUI.  It is
    # maintained by the undo manager.
    active_stack_clean = Bool

    # This is the name of the command that can be redone.  It will be empty if
    # there is no command that can be redone.  It is maintained by the undo
    # manager.
    redo_name = Unicode

    # This is the sequence number of the next command to be performed.  It is
    # incremented immediately before a command is invoked (by its 'do()'
    # method).
    sequence_nr = Int

    # This event is fired when the index of a command stack changes.  Note that
    # it may not be the active stack.
    stack_updated = Event(Instance('apptools.undo.api.ICommandStack'))

    # This is the name of the command that can be undone.  It will be empty if
    # there is no command that can be undone.  It is maintained by the undo
    # manager.
    undo_name = Unicode

    ###########################################################################
    # 'IUndoManager' interface.
    ###########################################################################

    def redo(self):
        """ Redo the last undone command of the active command stack. """

    def undo(self):
        """ Undo the last command of the active command stack. """



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" The base class for all context adapter factories. """


# Enthought library imports.
from apptools.type_manager.api import AdapterFactory

# Local imports.
from .context import Context


class ContextAdapterFactory(AdapterFactory):
    """ The base class for all context adapter factories. """

    #### 'AdapterFactory' interface ###########################################

    # The target class (the class that the factory can adapt objects to).
    target_class = Context

    ###########################################################################
    # Protected 'AbstractAdapterFactory' interface.
    ###########################################################################

    def _adapt(self, adaptee, target_class, environment, context):
        """ Returns an adapter that adapts an object to the target class. """

        adapter = self.adapter_class(
            adaptee=adaptee, environment=environment, context=context
        )

        return adapter

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought logger package component>
#------------------------------------------------------------------------------
""" Logger plugin.
"""

# Standard library imports.
import logging

# Enthought library imports.
from envisage.api import ExtensionPoint, Plugin
from apptools.logger.log_queue_handler import LogQueueHandler
from traits.api import Callable, List

# Local imports.
from .logger_preferences import LoggerPreferences
from .logger_service import LoggerService


ID = 'apptools.logger'
ILOGGER = ID + '.plugin.logger_service.LoggerService'

class LoggerPlugin(Plugin):
    """ Logger plugin.
    """

    id = ID
    name = 'Logger plugin'

    #### Extension points for this plugin ######################################

    MAIL_FILES = 'apptools.logger.plugin.mail_files'

    mail_files = ExtensionPoint(
        List(Callable), id=MAIL_FILES, desc="""

        This extension point allows you to contribute functions which will be
        called to add project files to the zip file that the user mails back
        with bug reports from the Quality Agent.

        The function will be passed a zipfile.ZipFile object.

        """
    )

    #### Contributions to extension points made by this plugin #################

    PREFERENCES = 'envisage.preferences'
    PREFERENCES_PAGES = 'envisage.ui.workbench.preferences_pages'
    VIEWS = 'envisage.ui.workbench.views'

    preferences = List(contributes_to=PREFERENCES)
    preferences_pages = List(contributes_to=PREFERENCES_PAGES)
    views = List(contributes_to=VIEWS)


    def _preferences_default(self):
        return ['pkgfile://%s/plugin/preferences.ini' % ID]

    def _preferences_pages_default(self):
        from apptools.logger.plugin.view.logger_preferences_page import \
            LoggerPreferencesPage
        return [LoggerPreferencesPage]

    def _views_default(self):
        return [self._logger_view_factory]


    #### Plugin interface ######################################################

    def start(self):
        """ Starts the plugin.
        """
        preferences = LoggerPreferences()
        service = LoggerService(application=self.application,
            preferences=preferences)
        formatter = logging.Formatter('%(levelname)s|%(asctime)s|%(message)s')
        handler = LogQueueHandler()
        handler.setLevel(preferences.level_)
        handler.setFormatter(formatter)
        root_logger = logging.getLogger()
        root_logger.addHandler(handler)
        root_logger.setLevel(preferences.level_)
        service.handler = handler
        self.application.register_service(ILOGGER, service)

    def stop(self):
        """ Stops the plugin.
        """
        service = self.application.get_service(ILOGGER)
        service.save_preferences()


    #### LoggerPlugin private interface ########################################

    def _logger_view_factory(self, **traits):
        from apptools.logger.plugin.view.logger_view import LoggerView
        service = self.application.get_service(ILOGGER)
        view = LoggerView(service=service, **traits)
        # Record the created view on the service.
        service.plugin_view = view
        return view


#### EOF ######################################################################



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2007 by Enthought, Inc.
#  All rights reserved.
#
#-----------------------------------------------------------------------------


"""
A re-usable method for calculating a unique name given a list of existing
names.

"""

def make_unique_name(base, existing=[], format="%s_%s"):
    """
    Return a name, unique within a context, based on the specified name.

    base: the desired base name of the generated unique name.
    existing: a sequence of the existing names to avoid returning.
    format: a formatting specification for how the name is made unique.

    """

    count = 2
    name = base
    while name in existing:
        name = format % (base, count)
        count += 1

    return name


#### EOF ####################################################################




""" Abstract base class for all object factories. """

# Standard library imports.
import logging

# Enthought library imports.
from traits.api import HasTraits


logger = logging.getLogger(__name__)


class AbstractFactory(HasTraits):
    """ Abstract base class for all object factories. """

    ###########################################################################
    # 'AbstractFactory' interface.
    ###########################################################################

    def create(self, target_class, *args, **kw):
        """ Creates an object of the specified target class.

        Returns None if the factory cannot produce an object of the target
        class.

        """

        if self._can_create(target_class, *args, **kw):
            obj = self._create(target_class, *args, **kw)
            if obj is None:
                logger.warn(self._get_warning_message(target_class))

        else:
            obj = None

        return obj

    ###########################################################################
    # Protected 'AbstractFactory' interface.
    ###########################################################################

    def _can_create(self, target_class, *args, **kw):
        """ Returns True if the factory can create objects of a class. """

        return NotImplementedError

    def _create(self, target_class, *args, **kw):
        """ Creates an object of the specified target class. """

        raise NotImplementedError

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_warning_message(self, target_class):
        """ Returns a warning message.

        The warning message is used when a factory fails to create something
        that it said it could!

        """

        message = "%s failed to create a %s" % (
            self.__class__.__name__,
            target_class.__name__
        )

        return message

#### EOF ######################################################################



def __replacement_setstate__(self, state):
    """
    """
    state = self.__updater__(state)
    self.__dict__.update(state)

    return




class Updater:

    """ An abstract class to provide functionality common to the updaters.
    """

    def get_latest(self, module, name):
        """ The refactorings dictionary contains mappings between old and new
        module names.  Since we only bump the version number one increment
        there is only one possible answer.
        """
        if hasattr(self, 'refactorings'):
            module = self.strip(module)
            name = self.strip(name)
            # returns the new module and name if it exists otherwise defaults
            # to using the original module and name
            module, name = self.refactorings.get((module, name), (module, name))

        return module, name


    def strip(self, string):
        # Who would have thought that pickle would pass us
        # names with \013 on the end? Is this after the files have
        # manually edited?
        if ord(string[-1:]) == 13:
            return string[:-1]

        return string

#### EOF #######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought application scripting package component>
#------------------------------------------------------------------------------


# Standard library imports.
import inspect
import types

# Enthought library imports.
from traits.api import HasTraits

# Local imports.
from .package_globals import get_script_manager
from .scriptable import scriptable, Scriptable


def create_scriptable_type(scripted_type, name=None, bind_policy='auto',
        api=None, includes=None, excludes=None, script_init=True):
    """Create and return a new type based on the given scripted_type that will
    (by default) have its public methods and traits (ie. those not beginning
    with an underscore) made scriptable.

    name is the name that objects of this type will be bound to.  It defaults
    to the name of scripted_type with the first character forced to lower case.
    It is ignored if script_init is False.

    bind_policy determines what happens if a name is already bound.  If the
    policy is 'auto' then a numerical suffix will be added to the name, if
    necessary, to make it unique.  If the policy is 'unique' then an exception
    is raised.  If the policy is 'rebind' then the previous binding is
    discarded.  It is ignored if script_init is False.  The default is 'auto'.

    If api is given then it is a class, or a list of classes, that define the
    attributes that will be made scriptable.

    Otherwise if includes is given it is a list of names of attributes that
    will be made scriptable.

    Otherwise all the public attributes of scripted_type will be made
    scriptable except those in the excludes list.

    Irrespective of any other arguments, if script_init is set then the
    __init__() method will always be made scriptable.
    """

    def __init__(self, *args, **kwargs):
        """Initialise the dynamic sub-class instance."""

        get_script_manager().new_object(self, scripted_type, args, kwargs,
                name, bind_policy)
        scripted_type.__init__(self, *args, **kwargs)

    # See if we need to pull all attribute names from a type.
    if api is not None:
        if isinstance(api, list):
            src = api
        else:
            src = [api]
    elif includes is None:
        src = [scripted_type]
    else:
        names = includes
        src = None

    if src:
        ndict = {}

        for cls in src:
            if issubclass(cls, HasTraits):
                for n in list(cls.class_traits().keys()):
                    if not n.startswith('_') and not n.startswith('trait'):
                        ndict[n] = None

            for c in inspect.getmro(cls):
                if c is HasTraits:
                    break

                for n in list(c.__dict__.keys()):
                    if not n.startswith('_'):
                        ndict[n] = None

        # Respect the excludes so long as there was no explicit API.
        if api is None and excludes is not None:
            for n in excludes:
                try:
                    del ndict[n]
                except KeyError:
                    pass

        names = list(ndict.keys())

    # Create the type dictionary containing replacements for everything that
    # needs to be scriptable.
    type_dict = {}
    if script_init:
        type_dict['__init__'] = __init__

    if issubclass(scripted_type, HasTraits):
        traits = scripted_type.class_traits()

        for n in names:
            trait = traits.get(n)

            if trait is not None:
                type_dict[n] = Scriptable(trait)

    for n in names:
        try:
            attr = getattr(scripted_type, n)
        except AttributeError:
            continue

        if type(attr) is types.MethodType:
            type_dict[n] = scriptable(attr)

    type_name = 'Scriptable(%s)' % scripted_type.__name__

    return type(type_name, (scripted_type, ), type_dict)


def make_object_scriptable(obj, api=None, includes=None, excludes=None):
    """Make (by default) an object's public methods and traits (ie. those not
    beginning with an underscore) scriptable.

    If api is given then it is a class, or a list of classes, that define the
    attributes that will be made scriptable.

    Otherwise if includes is given it is a list of names of attributes that
    will be made scriptable.

    Otherwise all the public attributes of scripted_type will be made
    scriptable except those in the excludes list.
    """

    # Create the new scriptable type.
    new_type = create_scriptable_type(obj.__class__, api=api,
            includes=includes, excludes=excludes, script_init=False)

    # Fix the object's type to make it scriptable.
    obj.__class__ = new_type



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" A null log handler. """


# Standard library imports.
import logging


class NullHandler(logging.Handler):
    """ A null log handler.

    This is a quick hack so that we can start to refactor the 'logger'
    module since it used to add actual handlers at module load time.

    Now we only add this handler so that people using just the ETS library and
    not one of our applications won't see the warning about 'no handlers'.

    """

    ###########################################################################
    # 'Handler' interface.
    ###########################################################################

    def emit(self, record):
        """ Emits a log record. """

        pass

#### EOF ######################################################################



import unittest

from ..type_registry import LazyRegistry

from .dummies import A, B, C, D, Mixed, Abstract, Concrete, ConcreteSubclass


class TestLazyRegistry(unittest.TestCase):
    def setUp(self):
        self.registry = LazyRegistry()
        # Bodge in a different importer.
        self.registry._import_object = 'Imported {0}'.format

    def test_deferred_push(self):
        self.registry.push('dummies:A', 'foo:A')
        self.registry.push('dummies:C', 'foo:C')
        self.assertEqual(self.registry.lookup_by_type(A), 'Imported foo:A')
        self.assertEqual(self.registry.lookup_by_type(B), 'Imported foo:A')
        self.assertEqual(self.registry.lookup_by_type(C), 'Imported foo:C')
        self.assertRaises(KeyError, self.registry.lookup_by_type, D)

    def test_greedy_push(self):
        self.registry.push(A, 'foo:A')
        self.registry.push(C, 'foo:C')
        self.assertEqual(self.registry.lookup_by_type(A), 'Imported foo:A')
        self.assertEqual(self.registry.lookup_by_type(B), 'Imported foo:A')
        self.assertEqual(self.registry.lookup_by_type(C), 'Imported foo:C')
        self.assertRaises(KeyError, self.registry.lookup_by_type, D)

    def test_mro(self):
        self.registry.push('dummies:A', 'foo:A')
        self.registry.push('dummies:D', 'foo:D')
        self.assertEqual(self.registry.lookup_by_type(Mixed), 'Imported foo:A')

    def test_lookup_instance(self):
        self.registry.push(A, 'foo:A')
        self.registry.push(C, 'foo:C')
        self.assertEqual(self.registry.lookup(A()), 'Imported foo:A')
        self.assertEqual(self.registry.lookup(B()), 'Imported foo:A')
        self.assertEqual(self.registry.lookup(C()), 'Imported foo:C')
        self.assertRaises(KeyError, self.registry.lookup, D())

    def test_abc(self):
        self.registry.push_abc(Abstract, 'foo:Abstract')
        self.assertEqual(self.registry.lookup_by_type(Concrete),
            'Imported foo:Abstract')
        self.assertEqual(self.registry.lookup_by_type(ConcreteSubclass),
            'Imported foo:Abstract')



#-------------------------------------------------------------------------------
#
#  An abstract base class implementation of the ITemplateDataNameItem interface
#  that looks for all specified values in its input context or optionally any of
#  its sub-contexts and outputs a context containing all such values found.
#
#  Written by: David C. Morrill
#
#  Date: 07/29/2007
#
#  (c) Copyright 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" An abstract base class implementation of the ITemplateDataNameItem interface
    that looks for all specified values in its input context or optionally any
    of its sub-contexts and outputs a context containing all such values found.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from traits.api \
    import HasPrivateTraits, Instance, Property, provides

from apptools.template.template_traits \
    import TBool

from apptools.template.itemplate_data_context \
    import ITemplateDataContext

from apptools.template.itemplate_data_name_item \
    import ITemplateDataNameItem

from apptools.template.template_impl \
    import Template

from .template_data_context \
    import TemplateDataContext

from .helper \
    import path_for

#-------------------------------------------------------------------------------
#  'AnyDataNameItem' class:
#-------------------------------------------------------------------------------

class AnyDataNameItem ( Template ):
    """ An abstract base class implementation of the ITemplateDataNameItem
        interface that looks for all specified values in its input context or
        optionally any of its sub-contexts and outputs a context containing all
        such values found.
    """

    implements ( ITemplateDataNameItem )

    #-- 'ITemplateDataNameItem' Interface Implementation -----------------------

    # The data context which this data name item should match against:
    input_data_context = Instance( ITemplateDataContext )

    # The data context containing the data values and/or contexts this data
    # name item matches:
    output_data_context = Instance( ITemplateDataContext )

    # The ITemplateChoice instance representing the current settings of the
    # data name item. This value must be read/write, and must be overridden by
    # sublasses.
    data_name_item_choice = Property

    # The alternative choices the user has for the data name item settings for
    # the current input data context. The list may be empty, in which case the
    # user cannot change the settings of the data name item. This value can be
    # read only, and must be overridden by subclasses.
    data_name_item_choices = Property

    #-- Public Traits ----------------------------------------------------------

    # Should all sub-contexts be included in the search:
    recursive = TBool( False )

    # Should included sub-contexts be flattened into a single context?
    flatten = TBool( False )

    #-- Private Traits ---------------------------------------------------------

    # The current recursive setting:
    current_recursive = TBool( False )

    # The current input data context:
    current_input_data_context = Property

    #-- Abstract Methods (Must be overridden in subclasses) --------------------

    def filter ( self, name, value ):
        """ Returns **True** if the specified context data *name* and *value*
            should be included in the output context; and **False** otherwise.
        """
        raise NotImplementedError

    #-- Property Implementations -----------------------------------------------

    def _get_data_name_item_choice ( self ):
        raise NotImplementedError

    def _set_data_name_item_choice ( self, value ):
        raise NotImplementedError

    def _get_data_name_item_choices ( self ):
        raise NotImplementedError

    def _get_current_input_data_context ( self ):
        return self.input_data_context

    #-- Trait Event Handlers ---------------------------------------------------

    def _recursive_changed ( self, value ):
        """ Handles the primary recursive setting being changed.
        """
        self.current_recursive = value

    def _input_data_context_changed ( self ):
        """ Handles the 'input_data_context' trait being changed.
        """
        self.inputs_changed()

    #-- Private Methods --------------------------------------------------------

    def inputs_changed ( self ):
        """ Handles any input value being changed. This method should be called
            by subclasses when any of their input values change.
        """
        output_context = None
        input_context  = self.current_input_data_context
        if input_context is not None:
            values = {}

            if self.current_recursive:
                if self.flatten:
                    self._add_context( input_context, values )
                else:
                    self._copy_context( input_context, values )
            else:
                self._add_values(  input_context, values, '' )

            if len( values ) > 0:
                output_context = TemplateDataContext(
                    data_context_path = input_context.data_context_path,
                    data_context_name = input_context.data_context_name,
                    values            = values )

        self.output_data_context = output_context

    def _add_values ( self, input_context, values, path = '' ):
        """ Adds all of the matching values in the specified *input_context* to
            the specified *values* dictionary.
        """
        # Filter each name/value in the current input context to see if it
        # should be added to the output values:
        filter = self.filter
        gdcv   = input_context.get_data_context_value
        for name in input_context.data_context_values:
            value = gdcv( name )
            if self.filter( name, value ):
                values[ path_for( path, name ) ] = value

    def _add_context ( self, input_context, values, path = '' ):
        """ Adds all of the matching values in the specified *input_context* to
            the specified *output_context*, and then applies itself recursively
            to all contexts contained in the specified *input_context*.
        """
        # Add all of the filtered values in the specified input context:
        self._add_values( input_context, values, path )

        # Now process all of the input context's sub-contexts:
        gdc = input_context.get_data_context
        for name in input_context.data_contexts:
            self._add_context( gdc( name ), values, path_for( path,
                                            input_context.data_context_name ) )

    def _copy_context ( self, input_context ):
        """ Clone the input context so that the result only contains values and
            contexts which contain valid values and are not empty.
        """
        values   = {}
        contexts = {}

        # Add all of the filtered values in the specified input context:
        self._add_values( input_context, values )

        # Now process all of the input context's sub-contexts:
        gdc = input_context.get_data_context
        for name in input_context.data_contexts:
            context = self._copy_context( gdc( name ) )
            if context is not None:
                contexts[ name ] = context

        if (len( values ) == 0) and (len( contexts ) == 0):
            return None

        return TemplateDataContext(
                    data_context_path = input_context.data_context_path,
                    data_context_name = input_context.data_context_name,
                    values            = values,
                    contexts          = contexts )




#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from pyface.action.api import Action
from traits.api import Bool, Event, Instance, Interface, List

# Local imports.
from .i_user import IUser


class IUserManager(Interface):
    """The interface implemented by a user manager to manage users."""

    # The list of PyFace management actions (ie. actions related to all users)
    # implemented by this user manager.
    management_actions = List(Instance(Action))

    # The current user.
    user = Instance(IUser)

    # The list of PyFace user actions (ie. actions related to the current user)
    # implemented by this user manager.
    user_actions = List(Instance(Action))

    # This is fired whenever the currently authenticated user changes.  It will
    # be None if the current user isn't authenticated.
    user_authenticated = Event(IUser)

    def bootstrapping(self):
        """Return True if the user manager is bootstrapping.  Typically this is
        when no users have been defined."""

    def authenticate_user(self):
        """Authenticate (ie. login) the user.  If successfully authenticated
        all secured objects are re-enabled according to the user's permissions.
        """

    def unauthenticate_user(self):
        """Unauthenticate (ie. logout) the user.  All secured objects are
        disabled."""

    def select_user(self, name):
        """Return an object that implements IUser for the user selected based
        on the given name.  If there was no user selected then return None.
        How the name is interpreted (eg. as a regular expression) is determined
        by the user manager."""



""" Utility functions.

fixme: I don't like random collections of utility functions! Where should
this go?

"""


# Standard library imports.
import os
from os.path import basename, dirname, isdir, splitdrive, splitext
from zipfile import is_zipfile, ZipFile


def get_module_name(filename):
    """ Get the fully qualified module name for a filename.

    For example, if the filename is

    /enthought/envisage/core/core_plugin_definition.py

    this method would return

    envisage.core.core_plugin_definition

    """

    if os.path.exists(filename):
        # Get the name of the module minus the '.py'
        module, ext = os.path.splitext(os.path.basename(filename))

        # Start with the actual module name.
        module_path = [module]

        # If the directory is a Python package then add it to the module path.
        #return self.is_folder and '__init__.py' in os.listdir(self.path)

        parent = dirname(filename)
        while isdir(parent) and '__init__.py' in os.listdir(parent):
            bname = basename(parent)
            module_path.insert(0, splitext(bname)[0])
            parent = dirname(parent)

        module_name = '.'.join(module_path)

    # If the file does not exist then it might be a zip file path.
    else:
        module_name = get_module_name_from_zip(filename)

    return module_name

# fixme: WIP
def get_module_name_from_zip(filename):

    # first, find the zip file in the path
    filepath = filename
    zippath = None
    while not is_zipfile(filepath) and \
              splitdrive(filepath)[1] != '\\' \
              and splitdrive(filepath)[1] != '/':
        filepath, tail = os.path.split(filepath)
        if zippath is not None:
            zippath = tail + '/' + zippath
        else:
            zippath = tail

    if not is_zipfile(filepath):
        return None

    # if the split left a preceding slash on the zippath then remove
    # it
    if zippath.startswith('\\') or zippath.startswith('/'):
        zippath = zippath[1:]

    # replace any backwards slashes with forward slashes
    zippath = zippath.replace('\\', '/')

    # Get the name of the module minus the '.py'
    module, ext = splitext(basename(zippath))

    # Start with the actual module name.
    module_path = [module]

    # to get the module name, we walk through the zippath until we
    # find a parent directory that does NOT have a __init__.py file
    z = ZipFile(filepath)

    parentpath = dirname(zippath)
    while path_exists_in_zip(z, parentpath + '/__init__.py'):
        module_path.insert(0, basename(parentpath))
        parentpath = dirname(parentpath)

    z.close()

    return '.'.join(module_path)

# fixme: WIP
def path_exists_in_zip(zfile, path):

    try:
        zfile.getinfo(path)
        exists = True
    except:
        exists = False

    return exists

# fixme: WIP
def is_zip_path(path):
    """ Returns True if the path refers to a zip file. """

    filepath = path
    while not is_zipfile(filepath) and \
              splitdrive(filepath)[1] != '\\' \
              and splitdrive(filepath)[1] != '/':
        filepath = dirname(filepath)

    return is_zipfile(filepath)

# fixme: WIP
def get_zip_path(filename):
    """ Returns the path to the zip file contained in the filename.

    fixme: An example here would help.

    """

    filepath = filename
    zippath = None
    while not is_zipfile(filepath) and \
              splitdrive(filepath)[1] != '\\' \
              and splitdrive(filepath)[1] != '/':
        filepath, tail = os.path.split(filepath)
        if zippath is not None:
            zippath = tail + '/' + zippath
        else:
            zippath = tail

    return zippath

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Major library imports.
from pyface.qt import QtGui

# Enthought library imports.
from apptools.permissions.adapter_base import AdapterBase


class QWidgetAdapter(AdapterBase):
    """This is the adapter for PyQt QWidget instances."""

    def adapt(self):
        """Reimplemented to adapt the proxied object."""

        # Replace the real methods.
        self.proxied.setEnabled = self.update_enabled
        self.proxied.setVisible = self.update_visible
        self.proxied.hide = self._hide
        self.proxied.show = self._show

        return self.proxied

    def get_enabled(self):
        """Get the enabled state of the proxied object."""

        return self.proxied.isEnabled()

    def set_enabled(self, value):
        """Set the enabled state of the proxied object."""

        QtGui.QWidget.setEnabled(self.proxied, value)

    def get_visible(self):
        """Get the visible state of the proxied object."""

        return self.proxied.isVisible()

    def set_visible(self, value):
        """Set the visible state of the proxied object."""

        QtGui.QWidget.setVisible(self.proxied, value)

    def _hide(self):
        """The replacement QWidget.hide() implementation."""

        self.update_visible(False)

    def _show(self):
        """The replacement QWidget.show() implementation."""

        self.update_visible(True)

AdapterBase.register_adapter(QWidgetAdapter, QtGui.QWidget)



# Copyright (c) 2007-2014 by Enthought, Inc.
# All rights reserved.

try:
    from apptools._version import full_version as __version__
except ImportError:
    __version__ = 'not-built'

__requires__ = [
    'traitsui',
    'configobj',
]



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from traits.api import HasTraits, Instance, provides

# Local imports.
from .i_policy_storage import IPolicyStorage, PolicyStorageError
from .persistent import Persistent, PersistentError


@provides(IPolicyStorage)
class PolicyStorage(HasTraits):
    """This implements a policy database that pickles its data in a local file.
    """



    #### Private interface ####################################################

    # The persisted database.  The database itself is a tuple of the role and
    # assignment dictionaries.  The role dictionary is keyed by the role name,
    # and has a value that is a tuple of the description and the list of
    # permission names.  The assigment dictionary is keyed by the user name and
    # has a value that is the list of role names.
    _db = Instance(Persistent)

    ###########################################################################
    # 'IPolicyStorage' interface.
    ###########################################################################

    def add_role(self, name, description, perm_ids):
        """Add a new role."""

        self._db.lock()

        try:
            roles, assigns = self._db.read()

            if name in roles:
                raise PolicyStorageError("The role \"%s\" already exists." % name)

            roles[name] = (description, perm_ids)
            self._db.write((roles, assigns))
        finally:
            self._db.unlock()

    def all_roles(self):
        """Return a list of all roles."""

        roles, _ = self._readonly_copy()

        return [(name, description) for name, (description, _) in list(roles.items())]

    def delete_role(self, name):
        """Delete a role."""

        self._db.lock()

        try:
            roles, assigns = self._db.read()

            if name not in roles:
                raise PolicyStorageError("The role \"%s\" does not exist." % name)

            del roles[name]

            # Remove the role from any users who have it.
            for user, role_names in list(assigns.items()):
                try:
                    role_names.remove(name)
                except ValueError:
                    continue

                assigns[user] = role_names

            self._db.write((roles, assigns))
        finally:
            self._db.unlock()

    def get_assignment(self, user_name):
        """Return the details of the assignment for the given user name."""

        _, assigns = self._readonly_copy()

        try:
            role_names = assigns[user_name]
        except KeyError:
            return '', []

        return user_name, role_names

    def get_policy(self, user_name):
        """Return the details of the policy for the given user name."""

        roles, assigns = self._readonly_copy()

        try:
            role_names = assigns[user_name]
        except KeyError:
            return '', []

        perm_ids = []

        for r in role_names:
            _, perms = roles[r]
            perm_ids.extend(perms)

        return user_name, perm_ids

    def is_empty(self):
        """See if the database is empty."""

        roles, assigns = self._readonly_copy()

        # Both have to be non-empty for the whole thing to be non-empty.
        return (len(roles) == 0 or len(assigns) == 0)

    def matching_roles(self, name):
        """Return the full name, description and permissions of all the roles
        that match the given name."""

        roles, _ = self._readonly_copy()

        # Return any role that starts with the name.
        roles = [(full_name, description, perm_ids)
                for full_name, (description, perm_ids) in list(roles.items())
                        if full_name.startswith(name)]

        return sorted(roles)

    def modify_role(self, name, description, perm_ids):
        """Update the description and permissions for the given role."""

        self._db.lock()

        try:
            roles, assigns = self._db.read()

            if name not in roles:
                raise PolicyStorageError("The role \"%s\" does not exist." % name)

            roles[name] = (description, perm_ids)
            self._db.write((roles, assigns))
        finally:
            self._db.unlock()

    def set_assignment(self, user_name, role_names):
        """Save the roles assigned to a user."""

        self._db.lock()

        try:
            roles, assigns = self._db.read()

            if len(role_names) == 0:
                # Delete the user, but don't worry if there is no current
                # assignment.
                try:
                    del assigns[user_name]
                except KeyError:
                    pass
            else:
                assigns[user_name] = role_names

            self._db.write((roles, assigns))
        finally:
            self._db.unlock()

    ###########################################################################
    # Trait handlers.
    ###########################################################################

    def __db_default(self):
        """Return the default persisted database."""

        return Persistent(self._db_factory, 'ets_perms_policydb',
                "the policy database")

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _readonly_copy(self):
        """Return the policy database (which should not be modified)."""

        try:
            self._db.lock()

            try:
                data = self._db.read()
            finally:
                self._db.unlock()
        except PersistentError as e:
            raise PolicyStorageError(str(e))

        return data

    @staticmethod
    def _db_factory():
        """Return an empty policy database."""

        return ({}, {})



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2006 by Enthought, Inc.
#  All rights reserved.
#
#-----------------------------------------------------------------------------

""" Provider of a framework that dynamically determines the contents of a
    context at the time of interaction with the contents rather than at the
    time a class is written.

    This capability is particularly useful when the object acting as a context
    is part of a plug-in application -- such as Envisage.  In general, this
    capability allows the context to be:

    - Extendable by contributions from somewhere other than the original
      code writer
    - Dynamic in that the elements it is composed of can change each time
      someone interacts with the contents of the context.

    It should be noted that this capability is explicitly different from
    contexts that look at another container to determine their contents, such
    as a file system context!

    Users of this framework contribute items to a dynamic context by adding
    traits to the dynamic context instance.  (This addition can happen
    statically through the use of a Traits Category.)  The trait value is the
    context item's value and the trait definition's metadata determines how the
    item is treated within the context.  The support metadata is:

    context_name: A non-empty string
        Represents the name of the item within this context.  This must be
        present for the trait to show up as a context item though the value
        may change over time as the item gets bound to different names.
    context_order: A float value
        Indicates the position for the item within this context.  All
        dynamically contributed context items are sorted by ascending order
        of this value using the standard list sort function.
    is_context: A boolean value
        True if the item is itself a context.
"""

# Standardlibrary imports
import logging

# Local imports
from .binding import Binding
from .context import Context
from .exception import OperationNotSupportedError


# Setup a logger for this module.
logger = logging.getLogger(__name__)


class DynamicContext(Context):
    """ A framework that dynamically determines the contents of a context at
        the time of interaction with the contents rather than at the time a
        context class is written.

        It should be noted that this capability is explicitly different from
        contexts that look at another container to determine their contents,
        such as a file system context!
    """

    ##########################################################################
    # 'Context' interface.
    ##########################################################################

    ### protected interface ##################################################

    def _is_bound(self, name):
        """ Is a name bound in this context?
        """

        item = self._get_contributed_context_item(name)
        result = item != (None, None)

        return result


    def _is_context(self, name):
        """ Returns True if a name is bound to a context.
        """
        item = self._get_contributed_context_item(name)
        if item != (None, None):
            obj, trait = item
            result = True == trait.is_context
        else:
            result = False

        return result


    def _list_bindings(self):
        """ Lists the bindings in this context.
        """
        result = [ Binding(name=n, obj=o, context=self) for n, o, t in \
                self._get_contributed_context_items() ]

        return result


    def _list_names(self):
        """ Lists the names bound in this context.
        """
        result = [ n for n, o, t in self._get_contributed_context_items() ]

        return result


    def _lookup(self, name):
        """ Looks up a name in this context.
        """
        item = self._get_contributed_context_item(name)
        if item != (None, None):
            obj, trait = item
            result = obj
        else:
            result = None

        return result


    def _rename(self, old_name, new_name):
        """ Renames an object in this context.
        """

        item = self._get_contributed_context_item(old_name)
        if item != (None, None):
            obj, trait = item
            trait.context_name = new_name
        else:
            raise ValueError('Name "%s" not in context', old_name)


    def _unbind(self, name):
        """ Unbinds a name from this context.
        """
        # It is an error to try to unbind any contributed context items
        item = self._get_contributed_context_item(name)
        if item != (None, None):
            raise OperationNotSupportedError('Unable to unbind ' + \
                'built-in with name [%s]' % name)


    ##########################################################################
    # 'DynamicContext' interface.
    ##########################################################################

    ### protected interface ##################################################

    def _get_contributed_context_item(self, name):
        """ If the specified name matches a contributed context item then
            returns a tuple of the item's current value and trait definition
            (in that order.)  Otherwise, returns a tuple of (None, None).
        """
        result = (None, None)

        for n, o, t in self._get_contributed_context_items():
            if n == name:
                result = (o, t)

        return result


    def _get_contributed_context_items(self):
        """ Returns an ordered list of items to be treated as part of our
            context.

            Each item in the list is a tuple of its name, object, and trait
            definition (in that order.)
        """
        # Our traits that get treated as context items are those that declare
        # themselves via metadata on the trait definition.
        filter = {
            'context_name': lambda v: v is not None and len(v) > 0
            }
        traits = self.traits(**filter)

        # Sort the list of context items according to the name of the item.
        traits = [ (t.context_order, n, t) for n, t in list(traits.items()) ]
        traits.sort()

        # Convert these trait definitions into a list of name and object tuples.
        result = [(t.context_name, getattr(self, n), t) for order, n, t \
            in traits]

        return result


### EOF ######################################################################




""" The help code implementation.

    This may be used to define examples to be displayed or demos to be run.

    :Copyright: 2008, Enthought Inc.
    :License: BSD
    :Author: Janet Swisher
"""
# This software is provided without warranty under the terms of the BSD
# license included in AppTools/trunk/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!

from apptools.preferences.api import PreferencesHelper
from traits.api import File, Str, provides

from .i_help_code import IHelpCode

@provides(IHelpCode)
class HelpCode(PreferencesHelper):
    """ The implementation for help codes.

    A help code is defined by a UI label and a filename.
    """


    #### IHelpCode interface / Preferences #####################################

    # NOTE: This class inherits preferences_page from PreferencesHelper.

    # The UI label for the help code, which appears in menus or dialogs.
    label = Str

    # The path to the entry point, which can be full, or relative to the Python
    # installation directory (sys.prefix).
    filename = File

    # The code to execute. This is executed when filename is None or an empty
    # string.
    code = Str




#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought undo package component>
#------------------------------------------------------------------------------



# Enthought library imports.
from traits.api import Bool, Instance, Interface, Unicode

# Local imports.
from .i_undo_manager import IUndoManager


class ICommandStack(Interface):
    """ The command stack interface.  A command stack is responsible for
    managing the changes to a data model and recording those changes so that
    they can be undone or redone.
    """

    #### 'ICommandStack' interface ############################################

    # This is the clean state of the stack.  Its value changes as commands are
    # undone and redone.  It can also be explicity set to mark the current
    # stack position as being clean (when the data is saved to disk for
    # example).
    clean = Bool

    # This is the name of the command that can be redone.  It will be empty if
    # there is no command that can be redone.  It is maintained by the undo
    # stack.
    redo_name = Unicode

    # This is the undo manager that manages this stack.
    undo_manager = Instance(IUndoManager)

    # This is the name of the command that can be undone.  It will be empty if
    # there is no command that can be undone.  It is maintained by the undo
    # stack.
    undo_name = Unicode

    ###########################################################################
    # 'ICommandStack' interface.
    ###########################################################################

    def begin_macro(self, name):
        """ This begins a macro by creating an empty command with the given
        'name'.  The commands passed to all subsequent calls to 'push()' will
        be contained in the macro until the next call to 'end_macro()'.  Macros
        may be nested.  The stack is disabled (ie. nothing can be undone or
        redone) while a macro is being created (ie. while there is an
        outstanding 'end_macro()' call).
        """

    def clear(self):
        """ This clears the stack, without undoing or redoing any commands, and
        leaves the stack in a clean state.  It is typically used when all
        changes to the data have been abandoned.
        """

    def end_macro(self):
        """ This ends a macro. """

    def push(self, command):
        """ This executes a command and saves it on the command stack so that
        it can be subsequently undone and redone.  'command' is an instance
        that implements the ICommand interface.  Its 'do()' method is called
        to execute the command.  If any value is returned by 'do()' then it is
        returned by 'push()'.  The command stack will keep a reference to the
        result so that it can recognise it as an argument to a subsequent
        command (which allows a script to properly save a result needed later).
        """

    def redo(self, sequence_nr=0):
        """ If 'sequence_nr' is 0 then the last command that was undone is
        redone and any result returned.  Otherwise commands are redone up to
        and including the given 'sequence_nr' and any result of the last of
        these is returned.
        """

    def undo(self, sequence_nr=0):
        """ If 'sequence_nr' is 0 then the last command is undone.  Otherwise
        commands are undone up to and including the given 'sequence_nr'.
        """



from .logger import add_log_queue_handler, create_log_file_handler
from .logger import FORMATTER, LEVEL, LogFileHandler
from .log_point import log_point
from .filtering_handler import FilteringHandler
from .null_handler import NullHandler



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought IO package component>
#------------------------------------------------------------------------------
""" Tests file operations. """


# Standard library imports.
import os, shutil, stat, unittest

# Enthought library imports.
from apptools.io import File


class FileTestCase(unittest.TestCase):
    """ Tests file operations on a local file system. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        try:
            shutil.rmtree('data')

        except:
            pass

        os.mkdir('data')

        return

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        shutil.rmtree('data')

        return

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_properties(self):
        """ file properties """

        # Properties of a non-existent file.
        f = File('data/bogus.xx')

        self.assert_(os.path.abspath(os.path.curdir) in f.absolute_path)
        self.assertEqual(f.children, None)
        self.assertEqual(f.ext, '.xx')
        self.assertEqual(f.exists, False)
        self.assertEqual(f.is_file, False)
        self.assertEqual(f.is_folder, False)
        self.assertEqual(f.is_package, False)
        self.assertEqual(f.is_readonly, False)
        self.assertEqual(f.mime_type, 'content/unknown')
        self.assertEqual(f.name, 'bogus')
        self.assertEqual(f.parent.path, 'data')
        self.assertEqual(f.path, 'data/bogus.xx')
        self.assert_(os.path.abspath(os.path.curdir) in f.url)
        self.assertEqual(str(f), 'File(%s)' % f.path)

        # Properties of an existing file.
        f = File('data/foo.txt')
        f.create_file()

        self.assert_(os.path.abspath(os.path.curdir) in f.absolute_path)
        self.assertEqual(f.children, None)
        self.assertEqual(f.ext, '.txt')
        self.assertEqual(f.exists, True)
        self.assertEqual(f.is_file, True)
        self.assertEqual(f.is_folder, False)
        self.assertEqual(f.is_package, False)
        self.assertEqual(f.is_readonly, False)
        self.assertEqual(f.mime_type, 'text/plain')
        self.assertEqual(f.name, 'foo')
        self.assertEqual(f.parent.path, 'data')
        self.assertEqual(f.path, 'data/foo.txt')
        self.assert_(os.path.abspath(os.path.curdir) in f.url)

        # Make it readonly.
        os.chmod(f.path, stat.S_IRUSR)
        self.assertEqual(f.is_readonly, True)

        # And then make it NOT readonly so that we can delete it at the end of
        # the test!
        os.chmod(f.path, stat.S_IRUSR | stat.S_IWUSR)
        self.assertEqual(f.is_readonly, False)

        return

    def test_copy(self):
        """ file copy """

        content = 'print "Hello World!"\n'

        f = File('data/foo.txt')
        self.assertEqual(f.exists, False)

        # Create the file.
        f.create_file(content)
        self.assertEqual(f.exists, True)
        self.assertRaises(ValueError, f.create_file, content)

        self.assertEqual(f.children, None)
        self.assertEqual(f.ext, '.txt')
        self.assertEqual(f.is_file, True)
        self.assertEqual(f.is_folder, False)
        self.assertEqual(f.mime_type, 'text/plain')
        self.assertEqual(f.name, 'foo')
        self.assertEqual(f.path, 'data/foo.txt')

        # Copy the file.
        g = File('data/bar.txt')
        self.assertEqual(g.exists, False)

        f.copy(g)
        self.assertEqual(g.exists, True)

        self.assertEqual(g.children, None)
        self.assertEqual(g.ext, '.txt')
        self.assertEqual(g.is_file, True)
        self.assertEqual(g.is_folder, False)
        self.assertEqual(g.mime_type, 'text/plain')
        self.assertEqual(g.name, 'bar')
        self.assertEqual(g.path, 'data/bar.txt')

        # Attempt to copy a non-existent file (should do nothing).
        f = File('data/bogus.xx')
        self.assertEqual(f.exists, False)

        g = File('data/bogus_copy.txt')
        self.assertEqual(g.exists, False)

        f.copy(g)
        self.assertEqual(g.exists, False)

        return

    def test_create_file(self):
        """ file creation """

        content = 'print "Hello World!"\n'

        f = File('data/foo.txt')
        self.assertEqual(f.exists, False)

        # Create the file.
        f.create_file(content)
        self.assertEqual(f.exists, True)
        self.assertEqual(open(f.path).read(), content)

        # Try to create it again.
        self.assertRaises(ValueError, f.create_file, content)

        return

    def test_delete(self):
        """ file deletion """

        content = 'print "Hello World!"\n'

        f = File('data/foo.txt')
        self.assertEqual(f.exists, False)

        # Create the file.
        f.create_file(content)
        self.assertEqual(f.exists, True)
        self.assertRaises(ValueError, f.create_file, content)

        self.assertEqual(f.children, None)
        self.assertEqual(f.ext, '.txt')
        self.assertEqual(f.is_file, True)
        self.assertEqual(f.is_folder, False)
        self.assertEqual(f.mime_type, 'text/plain')
        self.assertEqual(f.name, 'foo')
        self.assertEqual(f.path, 'data/foo.txt')

        # Delete it.
        f.delete()
        self.assertEqual(f.exists, False)

        # Attempt to delete a non-existet file (should do nothing).
        f = File('data/bogus.txt')
        self.assertEqual(f.exists, False)

        f.delete()
        self.assertEqual(f.exists, False)

        return

if __name__ == "__main__":
    unittest.main()

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" The base class for all directory contexts. """


# Enthought library imports.
from traits.api import Dict

# Local imports.
from .context import Context
from .exception import NameNotFoundError, NotContextError


class DirContext(Context):
    """ The base class for all directory contexts. """

    # The attributes of every object in the context.  The attributes for the
    # context itself have the empty string as the key.
    #
    # {str name : dict attributes}
    _attributes = Dict

    ###########################################################################
    # 'DirContext' interface.
    ###########################################################################

    def get_attributes(self, name):
        """ Returns the attributes associated with a named object. """

        # If the name is empty then we return the attributes of this context.
        if len(name) == 0:
            attributes = self._get_attributes(name)

        else:
            # Parse the name.
            components = self._parse_name(name)

            # If there is exactly one component in the name then the operation
            # takes place in this context.
            if len(components) == 1:
                atom = components[0]

                if not self._is_bound(atom):
                    raise NameNotFoundError(name)

                # Do the actual get.
                attributes = self._get_attributes(atom)

            # Otherwise, attempt to continue resolution into the next context.
            else:
                if not self._is_bound(components[0]):
                    raise NameNotFoundError(components[0])

                next_context = self._get_next_context(components[0])
                attributes = next_context.get_attributes(
                    '/'.join(components[1:]))

        return attributes

    def set_attributes(self, name, attributes):
        """ Sets the attributes associated with a named object. """

        # If the name is empty then we set the attributes of this context.
        if len(name) == 0:
            attributes = self._set_attributes(name, attributes)

        else:
            # Parse the name.
            components = self._parse_name(name)

            # If there is exactly one component in the name then the operation
            # takes place in this context.
            if len(components) == 1:
                atom = components[0]

                if not self._is_bound(atom):
                    raise NameNotFoundError(name)

                # Do the actual set.
                self._set_attributes(atom, attributes)

            # Otherwise, attempt to continue resolution into the next context.
            else:
                if not self._is_bound(components[0]):
                    raise NameNotFoundError(components[0])

                next_context = self._get_next_context(components[0])
                next_context.set_attributes(
                    '/'.join(components[1:]), attributes
                )

        return

    # fixme: Non-JNDI
    def find_bindings(self, visitor):
        """ Find bindings with attributes matching criteria in visitor.

        Visitor is a function that is passed the bindings for each level of the
        heirarchy and the attribute dictionary for those bindings.  The visitor
        examines the bindings and dictionary and returns the bindings it is
        interested in.

        """

        bindings = visitor(self.list_bindings(), self._attributes)

        # recursively check other sub contexts.
        for binding in self.list_bindings():
            obj = binding.obj
            if isinstance(obj, DirContext):
                bindings.extend(obj.find_bindings(visitor))

        return bindings

    ###########################################################################
    # Protected 'DirContext' interface.
    ###########################################################################

    def _get_attributes(self, name):
        """ Returns the attributes of an object in this context. """

        attributes = self._attributes.setdefault(name, {})

        return attributes.copy()

    def _set_attributes(self, name, attributes):
        """ Sets the attributes of an object in this context. """

        self._attributes[name] = attributes

        return

    ###########################################################################
    # Protected 'Context' interface.
    ###########################################################################

    def _unbind(self, name):
        """ Unbinds a name from this context. """

        super(DirContext, self)._unbind(name)

        if name in self._attributes:
            del self._attributes[name]

        return

    def _rename(self, old_name, new_name):
        """ Renames an object in this context. """

        super(DirContext, self)._rename(old_name, new_name)

        if old_name in self._attributes:
            self._attributes[new_name] = self._attributes[old_name]
            del self._attributes[old_name]

        return

    def _destroy_subcontext(self, name):
        """ Destroys a sub-context of this context. """

        super(DirContext, self)._destroy_subcontext(name)

        if name in self._attributes:
            del self._attributes[name]

        return

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought logger package component>
#------------------------------------------------------------------------------


# Standard library imports.
import logging
from traceback import format_exception



"""
    To catch exceptions with our own code this code needs to be added
    sys.excepthook = custom_excepthook
"""

def custom_excepthook(type, value, traceback):
    """ Pass on the exception to the logging system. """

    msg = 'Custom - Traceback (most recent call last):\n'
    list = format_exception(type, value, traceback)

    msg = "".join(list)

    # Try to find the module that the exception actually came from.
    name = getattr(traceback.tb_frame, 'f_globals', {}).get('__name__',
        __name__)
    logger = logging.getLogger(name)
    logger.error(msg)

    return


## EOF ##################################################################



""" The definition of an Envisage plugin for online help.

It assumes that the Workbench plugin is being used.

    :Copyright: 2008, Enthought Inc.
    :License: BSD
    :Author: Janet Swisher
"""
# This software is provided without warranty under the terms of the BSD
# license included in AppTools/trunk/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!

# Standard library imports.
import logging
import new

# Enthought libary imports
from envisage.api import Plugin, ExtensionPoint
from envisage.ui.action.api import ActionSet, Group, Menu
from traits.api import Instance, List, Str

# Local imports
from .help_code import HelpCode
from .help_doc import HelpDoc
from .i_help_code import IHelpCode
from .i_help_doc import IHelpDoc

# Logging.
logger = logging.getLogger(__name__)

HELP_MENU = 'MenuBar/Help'
DOCS_MENU = 'Documents'
DOCS_GROUP = 'DocsGroup'
DEMOS_MENU = 'Demos'
DEMOS_GROUP = 'DemosGroup'
EXAMPLES_MENU = 'Examples'
EXAMPLES_GROUP = 'ExamplesGroup'
DOWNLOADS_MENU = 'Downloads'

# This module's package.
PKG = '.'.join(__name__.split('.')[:-1])


################################################################################
# The plugin
################################################################################
class HelpPlugin(Plugin):
    """ A Help plugin for Envisage Workbench applications.

    This plugin displays items on the Workbench Help menu, based on
    contributions from itself or other plugins.
    """
    # IDs of extension points this plugin offers.
    HELP_DOCS = PKG + '.help_docs'
    HELP_DEMOS = PKG + '.help_demos'
    HELP_EXAMPLES = PKG + '.help_examples'
    HELP_DOWNLOADS = PKG + '.help_downloads'

    # IDs of extension points this plugin contributes to.
    WORKBENCH_ACTION_SETS='envisage.ui.workbench.action_sets'
    PREFERENCES       = 'envisage.preferences'
    PREFERENCES_PAGES = 'envisage.ui.workbench.preferences_pages'

    #### 'IPlugin' interface ##################################################

    # The plugin's unique identifier.
    id = 'apptools.help.help_plugin'

    # The plugin's name (suitable for displaying to the user).
    name = 'Help Plugin'

    #### public 'HelpPlugin' interface ########################################

    # List of menu names added to the main toolbar by this plugin
    # (These are placed before the 'Help' Menu).
    menus = List(Str)

    #### Extension points offered by this plugin ##############################

    help_docs = ExtensionPoint(
        List(IHelpDoc), id=HELP_DOCS, desc="""

        A help doc is defined by a preference node that specifies a UI label,
        a filename for the document, and a (path to a) viewer for viewing the
        document.

        Each contribution to this extension point must be an instance of a
        class that implements IHelpDoc. The easiest way to do this is to
        create an instance of  `apptools.help.help_plugin.api.HelpDoc`.

        So, to contribute a help doc:

        1. Create a preferences file for your plugin if it doesn't already
           have one. (Be sure to contribute your preferences file to the
           `envisage.preferences` extension point.)

        2. Define a unique "node" (section heading) in your preferences file
           for each document, and specify values for the 'label', 'viewer',
           and 'filename' settings. (Use 'browser' as the value of 'viewer' if
           the document can be viewed in a web browser.)

        3. For each document, contribute a HelpDoc to this extension point,
           and specify its *preferences_path* as the corresponding node
           name from your preferences file.
        """
    )

    help_demos = ExtensionPoint(
        List(IHelpCode), id=HELP_DEMOS, desc="""

        A help demo is a Python program that is run when it is selected from
        the Help>Demos submenu. It is defined by a preference node that
        specifies a UI label and a filename for the demo entry point.

        Each contribution to this extension point must be an instance of a
        class that implements IHelpCode. The easiest way to do this is to
        create an instance of  `apptools.help.help_plugin.api.HelpCode`.

        So, to contribute a help demo:

        1. Create a preferences file for your plugin if it doesn't already
           have one. (Be sure to contribute your preferences file to the
           `envisage.preferences` extension point.)

        2. Define a unique "node" (section heading) in your preferences file
           for each demo, and specify values for the 'label' and 'filename'
           settings. (Note that the same preferences section can be used for
           a help demo and a help example.)

        3. For each demo, contribute a HelpDemo to this extension point,
           and specify its *preferences_path* as the corresponding node
           name from your preferences file.
        """
    )

    help_examples = ExtensionPoint(
        List(IHelpCode), id=HELP_EXAMPLES, desc="""

        A help example is a Python file that is opened for viewing when it is
        selected from the Help>Examples submenu. It is defined by a preference
        node that specifies a UI label and a filename for the primary example
        file.

        Each contribution to this extension point must be an instance of a
        class that implements IHelpCode. The easiest way to do this is to
        create an instance of `apptools.help.help_plugin.api.HelpCode`.

        So, to contribute a help example:

        1. Create a preferences file for your plugin if it doesn't already
           have one. (Be sure to contribute your preferences file to the
           `envisage.preferences` extension point.)

        2. Define a unique "node" (section heading) in your preferences file
           for each example, and specify values for the 'label' and 'filename'
           settings. (Note that the same preferences section can be used for
           a help demo and a help example.)

        3. For each example, contribute a HelpCode to this extension point,
           and specify its *preferences_path* as the corresponding node
           name from your preferences file.
        """
    )

    help_downloads = ExtensionPoint(
        List(IHelpDoc), id=HELP_DOWNLOADS, desc="""

        A help download is a url that is opened in a browser for viewing when it
        is selected from the Help>Downloads submenu. It is defined by a
        preference node that specifies a UI label and a url for the download.

        Each contribution to this extension point must be an instance of a
        class that implements IHelpDoc, and has the url trait set to True. The
        easiest way to do this is to create an instance of
        `apptools.help.help_plugin.api.HelpDoc`.

        So, to contribute a help doc:

        1. Create a preferences file for your plugin if it doesn't already
           have one. (Be sure to contribute your preferences file to the
           `envisage.preferences` extension point.)

        2. Define a unique "node" (section heading) in your preferences file
           for each url, and specify values for the 'label' and 'filename'
           settings. Also set 'url' to True.

        3. For each document, contribute a HelpDoc to this extension point,
           and specify its *preferences_path* as the corresponding node
           name from your preferences file.
        """
    )

    # FIXME: Ideally, there should be an extension point to allow plugins to
    #        offer editors to display examples (e.g., TextEditorPlugin or
    #        RemoteEditorPlugin). For now, examples open in an external editor
    #        launched with subprocess.Popen. The user can set the editor
    #        command in the Examples preferences page.


    ###### Contributions to extension points made by this plugin ######

    help_action_sets = List(contributes_to=WORKBENCH_ACTION_SETS)
    def _help_action_sets_default(self):
        """ Returns a list containing an action set class whose **actions**
            correspond to the help docs in the help_docs extension point.
        """
        extension_point_mapping = {
                                DOCS_MENU: self.help_docs,
                                EXAMPLES_MENU: self.help_examples,
                                DEMOS_MENU: self.help_demos,
                                DOWNLOADS_MENU: self.help_downloads}

        # Construct traits for the action set
        ns = {'id': 'apptools.help.help_plugin.help_action_set',
              'name': 'Help Plugin ActionSet',
              'groups': [ Group( id=DOCS_GROUP, before='AboutGroup',
                                 path=HELP_MENU ) ]
              }

        for (menu_name, items) in list(extension_point_mapping.items()):
            if len(items) > 0:
                menu = Menu( name = menu_name,
                             class_name =
                             PKG + '.help_submenu_manager:%sMenuManager' %
                             menu_name )
                if menu_name in self.menus:
                    menu.path = 'MenuBar'
                    menu.before = 'Help'
                else:
                    menu.path = HELP_MENU
                    menu.group = DOCS_GROUP

                # Append the menu.
                ns.setdefault('menus', []).append(menu)

        return [new.classobj('SPLHelpActionSet', (ActionSet,), ns)]

    preferences = List(contributes_to=PREFERENCES)
    def _preferences_default(self):
        return ['pkgfile://%s/preferences.ini' % PKG]

    preferences_pages = List(contributes_to=PREFERENCES_PAGES)
    def _preferences_pages_default(self):
        from apptools.help.help_plugin.preferences_pages import \
            DocumentsPreferencesPage, DemosPreferencesPage, \
            ExamplesPreferencesPage, HelpDocPreferencesPage, \
            HelpDemoPreferencesPage, HelpExamplePreferencesPage
        pages = []
        if len(self.help_docs) > 0:
            pages.append(DocumentsPreferencesPage)
            pages.extend(
                [ new.classobj(doc.preferences_path + 'PreferencesPage',
                              (HelpDocPreferencesPage,),
                              {'preferences_path': doc.preferences_path},
                             ) for doc in self.help_docs
                ])
        if len(self.help_demos) > 0:
            pages.append(DemosPreferencesPage)
            pages.extend(
                [ new.classobj(demo.preferences_path + 'PreferencesPage',
                              (HelpDemoPreferencesPage,),
                              {'preferences_path': demo.preferences_path},
                             ) for demo in self.help_demos
                ])
        if len(self.help_examples) > 0:
            pages.append(ExamplesPreferencesPage)
            pages.extend(
                [ new.classobj(example.preferences_path + 'PreferencesPage',
                              (HelpExamplePreferencesPage,),
                              {'preferences_path': example.preferences_path},
                             ) for example in self.help_examples
                ])
        return pages

    #my_help_demos = List(contributes_to=HELP_DEMOS)
    #def _my_help_demos_default(self):
    #    return [HelpCode( preferences_path=PKG + '.TraitsDemo'),
    #           ]

    #my_help_examples = List(contributes_to=HELP_EXAMPLES)
    #def _my_help_examples_default(self):
    #    return [HelpCode( preferences_path=PKG + '.AcmeLab'),
    #           ]



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" The base class for all object factories. """


# Enthought library imports.
from traits.api import HasTraits


class ObjectFactory(HasTraits):
    """ The base class for all object factories.

    An object factory accepts some information about how to create an object
    (such as a reference) and returns an instance of that object.

    """

    ###########################################################################
    # 'ObjectFactory' interface.
    ###########################################################################

    def get_object_instance(self, state, name, context):
        """ Creates an object using the specified state information.

        Returns None if the factory cannot create the object (ie. it does not
        recognise the state passed to it).

        """

        raise NotImplementedError

### EOF #######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from pyface.api import error
from pyface.action.api import Action
from traits.api import Dict, HasTraits, provides, Instance, List

# Local imports.
from apptools.permissions.i_policy_manager import IPolicyManager
from apptools.permissions.permission import ManagePolicyPermission, Permission
from apptools.permissions.secure_proxy import SecureProxy
from .i_policy_storage import IPolicyStorage, PolicyStorageError
from .role_assignment import role_assignment
from .role_definition import role_definition


@provides(IPolicyManager)
class PolicyManager(HasTraits):
    """The default policy manager implementation.  This policy enforces the use
    of roles.  Permissions are associated with roles rather than directly with
    users.  Users are then associated with one or more roles."""



    #### 'IPolicyManager' interface ###########################################

    management_actions = List(Instance(Action))

    user_permissions = List(Instance(Permission))

    #### 'PolicyManager' interface ############################################

    # The dictionary of registered permissions keyed on the permission name.
    permissions = Dict

    # The policy data storage.
    policy_storage = Instance(IPolicyStorage)

    ###########################################################################
    # 'IPolicyManager' interface.
    ###########################################################################

    def bootstrapping(self):
        """Return True if we are bootstrapping, ie. no roles have been defined
        or assigned."""

        try:
            bootstrap = self.policy_storage.is_empty()
        except PolicyStorageError:
            # Suppress the error and assume it isn't empty.
            bootstrap = False

        return bootstrap

    def load_policy(self, user):
        """Load the policy for the given user."""

        self.user_permissions = []

        # See if the policy is to be unloaded.
        if user is None:
            return

        # Get the user's policy.
        try:
            user_name, perm_ids = self.policy_storage.get_policy(user.name)
        except PolicyStorageError as e:
            error(None, str(e))
            return

        for id in perm_ids:
            try:
                permission = self.permissions[id]
            except KeyError:
                # This shouldn't happen if referential integrity is maintained.
                continue

            self.user_permissions.append(permission)

    def register_permission(self, permission):
        """Register the given permission."""

        if permission.id in self.permissions:
            other = self.permissions[permission.id]

            if other.application_defined:
                if permission.application_defined:
                    raise KeyError('permission "%s" has already been defined' % permission.id)

                # Use the description from the policy manager, if there is
                # one, in preference to the application supplied one.
                if permission.description:
                    other.description = permission.description
            elif permission.application_defined:
                # Again, prefer the policy manager description.
                if other.description:
                    permission.description = other.description

                self.permissions[permission.id] = permission
            else:
                # This should never happen if the policy manager is working
                # properly.
                raise KeyError('permission "%s" has already been defined by the same policy manager' % permission.id)
        else:
            self.permissions[permission.id] = permission

    ###########################################################################
    # Trait handlers.
    ###########################################################################

    def _management_actions_default(self):
        """Return the management actions to manage the policy."""

        actions = []
        perm = ManagePolicyPermission()

        act = Action(name='&Role Definitions...', on_perform=role_definition)
        actions.append(SecureProxy(act, permissions=[perm], show=False))

        act = Action(name='&Role Assignments...', on_perform=role_assignment)
        actions.append(SecureProxy(act, permissions=[perm], show=False))

        return actions

    def _policy_storage_default(self):
        """Return the default storage for the policy data."""

        # Defer to an external storage manager if there is one.
        try:
            from apptools.permissions.external.policy_storage import PolicyStorage
        except ImportError:
            from apptools.permissions.default.policy_storage import PolicyStorage

        return PolicyStorage()



import numpy as np
from pandas import DataFrame
from tables.table import Table as PyTablesTable


class _TableRowAccessor(object):
    """ A simple object which provides read access to the rows in a Table.
    """
    def __init__(self, h5_table):
        self._h5_table = h5_table

    def __getitem__(self, key):
        return self._h5_table[key]


class H5TableNode(object):
    """ A wrapper for PyTables Table nodes.

    Parameters
    ----------
    node : tables.Table instance
        An H5 node which is a pytables.Table or H5TableNode instance
    """

    def __init__(self, node):
        # Avoid a circular import
        from .file import H5Attrs

        assert self.is_table_node(node)
        self._h5_table = node._h5_table if hasattr(node, '_h5_table') else node
        self.attrs = H5Attrs(self._h5_table._v_attrs)

    #--------------------------------------------------------------------------
    #  Creation methods
    #--------------------------------------------------------------------------

    @classmethod
    def add_to_h5file(cls, h5, node_path, description, **kwargs):
        """ Add table node to an H5 file at the specified path.

        Parameters
        ----------
        h5 : H5File
            The H5 file where the table node will be stored.
        node_path : str
            Path to node where data is stored (e.g. '/path/to/my_table')
        description : list of tuples or numpy dtype object
            The description of the columns in the table. This is either a list
            of (column name, dtype, [, shape or itemsize]) tuples or a numpy
            record array dtype. For more information, see the documentation for
            `Table` in PyTables.
        **kwargs : dict
            Additional keyword arguments to pass to pytables.File.create_table
        """
        if isinstance(description, (tuple, list)):
            description = np.dtype(description)

        cls._create_pytables_node(h5, node_path, description, **kwargs)
        node = h5[node_path]

        return cls(node)

    @classmethod
    def is_table_node(cls, pytables_node):
        """ Return True if pytables_node is a pytables.Table or a H5TableNode.
        """
        return isinstance(pytables_node, (PyTablesTable, H5TableNode))

    #--------------------------------------------------------------------------
    #  Public interface
    #--------------------------------------------------------------------------

    def append(self, data):
        """ Add some data to the table.

        Parameters
        ----------
        data : dict
            A dictionary of column name -> values items
        """
        rows = list(zip(*[data[name] for name in list(self.keys())]))
        self._h5_table.append(rows)

    def __getitem__(self, col_or_cols):
        """ Return one or more columns of data from the table.

        Parameters
        ----------
        col_or_cols : str or list of str
            A single column name or a list of column names

        Return
        ------
        data : ndarray
            An array of column data with the column order matching that of
            `col_or_cols`.
        """
        if isinstance(col_or_cols, str):
            return self._h5_table.col(col_or_cols)

        column_data = [self._h5_table.col(name) for name in col_or_cols]
        return np.column_stack(column_data)

    @property
    def ix(self):
        """ Return an object which provides access to row data.
        """
        return _TableRowAccessor(self._h5_table)

    def keys(self):
        return self._h5_table.colnames

    def to_dataframe(self):
        """ Return table data as a pandas `DataFrame`.

        XXX: This does not work if the table contains a multidimensional column
        """
        # Slicing rows gives a numpy struct array, which DataFrame understands.
        return DataFrame(self.ix[:])

    #--------------------------------------------------------------------------
    #  Object interface
    #--------------------------------------------------------------------------

    def __repr__(self):
        return repr(self._h5_table)

    def __len__(self):
        return self._h5_table.nrows

    #--------------------------------------------------------------------------
    #  Private interface
    #--------------------------------------------------------------------------

    def _f_remove(self):
        """ Implement the PyTables `Node._f_remove` method so that H5File
        doesn't choke when trying to remove our node.
        """
        self._h5_table._f_remove()
        self._h5_table = None

    @classmethod
    def _create_pytables_node(cls, h5, node_path, description, **kwargs):
        path, name = h5.split_path(node_path)
        pyt_file = h5._h5
        pyt_file.create_table(path, name, description, **kwargs)
